from protosearch.utils.standards import CrystalStandards
from protosearch.workflow.workflow import Workflow

standard_lattice = CrystalStandards.standard_lattice_mp
all_elements = list(standard_lattice.keys())


class StandardStates:

    def __init__(self):
        pass

    def submit_standard_states(self, elements=None):
        elements = elements or all_elements
        W = Workflow()
        print(elements)
        for e in elements:
            print(e)
            prototype = standard_lattice[e]
            W.submit(prototype)



import time
import string
import numpy as np
import pandas as pd
from ase.symbols import string2symbols
from catlearn.active_learning import acquisition_functions
from catlearn.fingerprint.voro import VoronoiFingerprintGenerator

from protosearch.build_bulk.enumeration import (
    Enumeration, AtomsEnumeration, get_stoich_from_formulas, get_formulas)
from protosearch.workflow.prototype_db import PrototypeSQL
from protosearch.workflow.workflow import Workflow as WORKFLOW
from protosearch.ml_modelling.fingerprint import clean_features
from protosearch.ml_modelling.regression_model import get_regression_model
from protosearch.utils.standards import CrystalStandards
from protosearch.build_bulk.oqmd_interface import OqmdInterface


class ActiveLearningLoop:

    def __init__(self,
                 chemical_formulas,
                 elements=None,
                 structure_source=['prototypes', 'icsd'],
                 batch_size=10,
                 min_atoms=None,
                 max_atoms=None,
                 min_wyckoff=None,
                 max_wyckoff=None,
                 enumeration_type='wyckoff',
                 check_frequency=300,
                 frac_jobs_limit=0.7,
                 spacegroups=None,
                 workflow=WORKFLOW
                 ):
        """
        Module to run active learning loop

        Parameters:
        ----------
        Workflow:
            Workflow class to use
        chemical_formulas: list of strings
            chemical formulas to investigate, such as:
            ['IrO2', 'IrO3']
        structure_source: list
            Specify how to generate the structures. Posible list entries are:
            'icsd': Experimental OQMD entries only
            'prototypes': Enumerate all prototypes
        batch_size: int
            number of DFT jobs to submit simultaneously.
        check_frequency: float
            Frequency in (s) that the AL checks on the job state
        frac_jobs_limit: float
            Upper limit on the fraction of of jobs to be processed before
            stoping the loop
        max_atoms: int
            Max number of atoms to allow in the unit cell
        """
        if isinstance(chemical_formulas, str):
            chemical_formulas = [chemical_formulas]
        self.source = structure_source
        self.batch_size = batch_size
        self.min_atoms = min_atoms
        self.max_atoms = max_atoms
        self.min_wyckoff = min_wyckoff
        self.max_wyckoff = max_wyckoff
        self.enumeration_type = enumeration_type
        self.check_frequency = check_frequency
        self.frac_jobs_limit = frac_jobs_limit
        self.spacegroups = spacegroups

        self.set_formulas_and_elements(chemical_formulas, elements)

        self.db_filename = '_'.join(chemical_formulas) + '.db'
        self.Workflow = workflow(db_filename=self.db_filename)

        self.Workflow.write_status(
            chemical_formulas=chemical_formulas, batch_size=batch_size)

    def _initialize(self):
        if not self.status['enumerated']:
            self.enumerate_structures(self.spacegroups)
            self.Workflow.write_status(enumerated=1)
        if not self.status['fingerprinted']:
            self.generate_fingerprints()
            self.Workflow.write_status(fingerprinted=1)

        self.batch_no = (self.status.get('last_batch_no', None) or 0) + 1

        elements = []
        for element_list in self.elements.values():
            elements += element_list
        elements = list(set(elements))

        self.Workflow.write_status(last_batch_no=self.batch_no)

        failed_ids = self.monitor_submissions(batch_size=0,
                                              standard_state=1)
        if len(failed_ids) > 0:
            raise RuntimeError(
                'Standard state calculation failed for one or more species')

        self.Workflow.write_status(initialized=1)

    def set_formulas_and_elements(self, formulas, elements=None):
        alphabet = list(string.ascii_uppercase)

        if not formulas[0][0] == 'A':
            self.stoichiometries, self.elements =\
                get_stoich_from_formulas(formulas)
            return

        self.stoichiometries = []
        for formula in formulas:
            stoich = []
            for f in formula:
                if f.isdigit():
                    stoich = stoich[:-1] + [f]
                else:
                    stoich += ['1']
            self.stoichiometries += ['_'.join(stoich)]

        self.elements = elements

    def monitor_submissions(self, batch_size, **kwargs):
        WF = self.Workflow
        completed_ids, failed_ids, running_ids =\
            WF.check_submissions(**kwargs)
        t0 = time.time()
        min_running_jobs = batch_size
        while len(running_ids) > min_running_jobs:
            WF.recollect()
            temp_completed_ids, temp_failed_ids, running_ids =\
                WF.check_submissions(**kwargs)
            failed_ids += temp_failed_ids
            completed_ids += temp_completed_ids
            t = time.time() - t0

            print('Status for calculations at {:.2f} min:'
                  .format(t / 60))
            print('  {} completed'.format(len(completed_ids)))
            print('  {} running'.format(len(running_ids)))
            print('  {} errored'.format(len(failed_ids)))

            if temp_failed_ids:
                print('\nOne or more jobs failed:')
                for i in failed_ids:
                    row = self.Workflow.ase_db.get(id=i)
                    message = row.data.get('error', 'No message').split('\n')
                    if len(message) > 1:
                        message = message[-2]
                    else:
                        message = message[0]
                    print('  Job {}:'.format(i))
                    print('    ' + message)
            time.sleep(self.check_frequency)

        # Make sure the number or running jobs stays constant
        self.corrected_batch_size = min_running_jobs + \
            self.batch_size - len(running_ids)

        self.save_formation_energies()

        return failed_ids

    def run(self):
        """Run actice learning loop"""

        WF = self.Workflow
        self.status = self.Workflow.get_status()
        if not self.status['initialized']:
            self._initialize()
            self.batch_no += 1
            self.acquire_batch(method='random', batch_size=self.batch_size)
            WF.submit_id_batch(self.batch_ids)
        else:
            self.batch_no = self.status['last_batch_no']

        happy = False
        while not happy:
            self.monitor_submissions(batch_size=self.batch_size)
            self.Workflow.write_job_status()
            self.generate_fingerprints(completed=True)

            happy = self.evaluate()

            self.train_ids = self.Workflow.get_completed_structure_ids()
            self.test_ids = self.Workflow.get_uncompleted_structure_ids()
            self.get_ml_prediction()

            all_ids = np.append(self.test_ids, self.train_ids)
            all_energies = np.append(self.energies, self.targets)
            all_uncertainties = np.append(self.uncertainties,
                                          np.zeros_like(self.targets))
            self.Workflow.write_predictions(
                self.batch_no, all_ids, all_energies, all_uncertainties)
            self.acquire_batch()

            self.batch_no += 1

            WF.submit_id_batch(self.batch_ids)

    def test_run(self, acquire='UCB', kappa=None):
        """Use already completed calculations to test the loop """
        WF = self.Workflow
        self.status = self.Workflow.get_status()
        self.corrected_batch_size = self.batch_size

        self.batch_no = 1
        self.batch_ids = []
        self.test_ids = self.Workflow.get_initial_structure_ids(completed=True)

        completed_ids = np.array(self.Workflow.get_completed_structure_ids())

        indices = np.random.randint(len(completed_ids),
                                    size=self.batch_size)

        self.train_ids = completed_ids[indices]

        completed_initial = [self.Workflow.ase_db.get(
            int(t)).initial_id for t in self.train_ids]

        self.test_ids = np.array([
            t for t in self.test_ids if not t in completed_initial])

        while len(self.test_ids) > 0:
            self.get_ml_prediction()
            all_ids = np.append(self.test_ids, self.train_ids)
            all_energies = np.append(self.energies, self.targets)
            all_uncertainties = np.append(self.uncertainties,
                                          np.zeros_like(self.targets))

            prediction = {'batch_no': self.batch_no,
                          'ids': all_ids,
                          'energies': all_energies,
                          'vars': all_uncertainties}

            yield prediction

            if acquire == 'random':
                indices = np.random.randint(len(self.test_ids),
                                            size=self.batch_size)
                self.batch_ids = np.array(self.test_ids)[indices]
            else:
                self.acquire_batch(method=acquire, kappa=kappa)

            completed_ids = [self.Workflow.ase_db.get(int(t)).final_id
                             for t in self.batch_ids]
            self.train_ids = np.sort(np.append(self.train_ids, completed_ids))

            self.test_ids = np.sort(np.array([
                t for t in self.test_ids if not t in self.batch_ids]))

            self.batch_no += 1

    def get_frac_of_systems_processed(self):
        """
        Get the fraction of structures that have been processed, including
        completed as well as errored jobs.
        """
        n_nonrelaxed = self.Workflow.ase_db.count(relaxed=0)
        n_errored = self.Workflow.ase_db.count(relaxed=0, completed=-1)
        n_relaxed = self.Workflow.ase_db.count(relaxed=1)

        frac = (n_relaxed + n_errored) / n_nonrelaxed

        return frac

    def evaluate(self):
        happy = False

        # Upper limit on Systems/Jobs processed
        frac_i = self.get_frac_of_systems_processed()
        print("Structures processed: {:.2f} %".format(frac_i))
        if frac_i >= self.frac_jobs_limit:
            print("Ending Loop! Upper limit of processed jobs reached")
            happy = True

        return happy

    def enumerate_structures(self, spacegroups=None):

        if 'icsd' in self.source:
            # First take experimental structures
            OI = OqmdInterface()
            print('Enumerating experimental prototypes')

            chemical_formulas = get_formulas(
                self.elements, self.stoichiometries)

            for chemical_formula in chemical_formulas:
                OI.store_enumeration(filename=self.db_filename,
                                     chemical_formula=chemical_formula,
                                     max_atoms=self.max_atoms)

        # Enumerate prototypes as a next step
        # experimental structures will not be overwritten.
        if 'prototypes' in self.source:
            print('Enumerating Prototypes')
            for stoichiometry in self.stoichiometries:
                self.enumerate_prototypes(stoichiometry,
                                          spacegroups)

            AE = AtomsEnumeration(self.elements,
                                  self.max_atoms,
                                  self.spacegroups)
            print('Enumerating atoms')
            AE.store_atom_enumeration(filename=self.db_filename,
                                      multithread=False)

    def enumerate_prototypes(self, stoichiometry, spacegroups=None,
                             num_type='wyckoff'):
        if spacegroups is not None:
            SG_start_end = [[s, s] for s in spacegroups]
        else:
            SG_start_end = [[1, 230]]

        if self.enumeration_type == 'wyckoff':
            num_min = self.min_wyckoff
            num_max = self.max_wyckoff
        else:
            num_min = self.min_atoms
            num_max = self.max_atoms

        for SG_start, SG_end in SG_start_end:
            E = Enumeration(stoichiometry,
                            num_start=num_min,
                            num_end=num_max,
                            SG_start=SG_start,
                            SG_end=SG_end,
                            num_type=self.enumeration_type)
            E.store_enumeration(filename=self.db_filename)

    def expand_structures(self, chemical_formula=None, max_atoms=None):
        """Add additional structures to enumerated space, by specifying new chemical
        formula or allowing more atoms per unit cell"""
        if chemical_formula:
            self.chemical_fomulas += [chemical_formula]
        stoichiometries, elements =\
            get_stoich_from_formulas(self.chemical_formulas)
        if max_atoms:
            self.max_atoms = max_atoms

        for stoichiometry in stoichiometries:
            npoints = sum([int(s) for s in stoichiometry.split('_')])
            E = Enumeration(stoichiometry, num_start=1, num_end=npoints,
                            SG_start=1, SG_end=230, num_type='wyckoff')

            E.store_enumeration(filename=self.db_filename)

        AE = AtomsEnumeration(elements, self.max_atoms)
        AE.store_atom_enumeration(filename=self.db_filename,
                                  multithread=True)

        self.generate_fingerprints()

    def generate_fingerprints(self,
                              feature_methods=['voronoi'],
                              completed=False):
        self.Workflow._connect()
        ase_db = self.Workflow.ase_db
        atoms_list = []
        target_list = {}

        ids = self.Workflow.get_new_fingerprint_ids(completed=completed)
        if not ids:
            print('Fingerprints up do date!')
            return

        targets_df = pd.DataFrame(columns=['id', 'Ef'])

        for id in ids:
            row = ase_db.get(id=id)
            atoms_list += [row.toatoms()]
            if row.get('Ef', None) is not None:
                targets_df = targets_df.append({'id': id, 'Ef': row.Ef},
                                               ignore_index=True)

        targets_df = targets_df.astype({'id': int, 'Ef': float})

        # Catlearn Magpie interface
        VF = VoronoiFingerprintGenerator(atoms_list)
        fingerprints_df = VF.generate().assign(id=ids)
        fingerprints_df = fingerprints_df.astype({'id': int})

        self.Workflow.write_dataframe(table='fingerprint',
                                      df=fingerprints_df)

        self.Workflow.write_dataframe(table='target',
                                      df=targets_df)

    def save_formation_energies(self):
        ase_db = self.Workflow.ase_db

        for row in self.Workflow.ase_db.select('-Ef', relaxed=1):
            energy = row.energy
            elements, counts = np.unique(
                row.toatoms().symbols, return_counts=True)
            ref_energies = []
            for i, e in enumerate(elements):
                ref_row = list(self.Workflow.ase_db.select(e,
                                                           relaxed=1,
                                                           standard_state=1,
                                                           limit=1))
                if len(ref_row) == 0:
                    print('Standard state for {} not found'.format(row.formula))
                    if e == 'O':
                        ref_energies += [counts[i] * 4.66]
                else:
                    energy_atom = ref_row[0].energy / ref_row[0].natoms
                    ref_energies += [counts[i] * energy_atom]

            formation_energy = (energy - sum(ref_energies)) / row.natoms
            ase_db.update(id=row.id, Ef=formation_energy)

    def get_features(self, scale=False):
        train_features = self.Workflow.load_dataframe(
            'fingerprint', ids=self.train_ids)

        ids_train = train_features.pop('id')
        targets = self.Workflow.load_dataframe('target', ids=self.train_ids)
        ids_targets = targets.pop('id')
        test_features = self.Workflow.load_dataframe(
            'fingerprint', ids=self.test_ids)
        ids_test = test_features.pop('id')

        assert np.all(ids_train == ids_targets)
        assert np.all(ids_test == self.test_ids)
        self.targets = targets.values

        features, bad_indices = \
            clean_features({'train': train_features.values,
                            'test': test_features.values},
                           scale=scale)

        self.train_ids = np.delete(self.train_ids, bad_indices['train'])
        self.targets = np.delete(self.targets, bad_indices['train'])
        self.test_ids = np.delete(self.test_ids, bad_indices['test'])

        return features

    def get_ml_prediction(self, model='catlearn'):
        features = self.get_features()

        Model = get_regression_model(model)(features['train'],
                                            self.targets)

        predictions = Model.predict(features['test'])

        self.energies = predictions['prediction']
        index = [i for i in range(len(self.test_ids))
                 if np.isfinite(self.energies[i])]
        self.test_ids = np.array(self.test_ids)[index]
        self.energies = self.energies[index]
        self.uncertainties = predictions['uncertainty'][index]

    def acquire_batch(self, method='UCB', kappa=0.5, batch_size=None):
        if not batch_size:
            batch_size = self.corrected_batch_size

        if method == 'random':
            uncompleted_ids = self.Workflow.get_uncompleted_structure_ids()
            acquistision_values = np.zeros(len(uncompleted_ids))
            indices = np.random.randint(len(uncompleted_ids), size=batch_size)
            self.batch_ids = np.array(uncompleted_ids)[indices]
            return

        if method == 'UCB':
            """
            Upper confidence bound:
            Trade-off between low energy and high uncertainty
            """
            acquisition_values = self.energies - kappa * self.uncertainties
            indices = np.argsort(acquisition_values)
        elif method in ['optimistic', 'proximity', 'optimistic_proximity',
                        'probability_density']:
            AF = getattr(acquisition_functions, method)
            acquisition_values = AF(y_best=-np.min(self.targets),
                                    predictions=-self.energies,
                                    uncertainty=self.uncertainties)

            indices = np.argsort(acquisition_values)[::-1]

        elif method in ['EI', 'PI']:
            AF = getattr(acquisition_functions, method)
            acquisition_values = AF(y_best=np.min(self.targets),
                                    predictions=self.energies,
                                    uncertainty=self.uncertainties,
                                    objective='min'
                                    )

            indices = np.argsort(acquisition_values)[::-1]

        elif method == 'mix':
            """
            Choose two different batches - one targeting low energy,
            the other targeting high uncertainty
            """
            n_high_u = batch_size // 3
            n_high_e = batch_size - n_high_u
            indices_energy = np.argsort(self.energies)
            indices_u = np.argsort(-self.uncertainties)
            indices = np.append(indices_energy[:n_high_e],
                                indices_u[:n_high_u])
            acquisition_values = -self.uncertainties

        self.batch_ids = np.array(self.test_ids)[indices][
            :batch_size]
        self.acquisition_values = acquisition_values


if __name__ == "__main__":
    ALL = ActiveLearningLoop(
        chemical_formulas=['Cu2O'],
        max_atoms=10,
        batch_size=1,
        check_frequency=0.6,
        frac_jobs_limit=0.4)
    ALL.run()



import numpy as np
import scipy
from shapely.geometry import Polygon
from ase.data import covalent_radii as cradii
from pymatgen.io.ase import AseAtomsAdaptor
from pymatgen.analysis.ewald import EwaldSummation

from protosearch.utils.data import metal_numbers, prefered_O_state,\
    favored_O_connections, electronegs


fixed_oxi_states = {'O': -2,
                    'S': -2,
                    'N': -3,
                    'P': -3,
                    'F': -1,
                    'Cl': -1}


def expand_cell(atoms, cutoff=None, padding=None):
    """
    Copy from Catkit connectivity utils (written by Jacob Boes)
    Return Cartesian coordinates atoms within a supercell
    which contains repetitions of the unit cell which contains
    at least one neighboring atom.

    Parameters
    ----------
    atoms : Atoms object
        Atoms with the periodic boundary conditions and unit cell
        information to use.
    cutoff : float
        Radius of maximum atomic bond distance to consider.
    padding : ndarray (3,)
        Padding of repetition of the unit cell in the x, y, z
        directions. e.g. [1, 0, 1].

    Returns
    -------
    index : ndarray (N,)
        Indices associated with the original unit cell positions.
    coords : ndarray (N, 3)
        Cartesian coordinates associated with positions in the
        supercell.
    offsets : ndarray (M, 3)
        Integer offsets of each unit cell.
    """
    cell = atoms.cell
    pbc = atoms.pbc
    pos = atoms.positions

    if padding is None and cutoff is None:
        diags = np.sqrt((
            np.dot([[1, 1, 1],
                    [-1, 1, 1],
                    [1, -1, 1],
                    [-1, -1, 1]],
                   cell)**2).sum(1))

        if pos.shape[0] == 1:
            cutoff = max(diags) / 2.
        else:
            dpos = (pos - pos[:, None]).reshape(-1, 3)
            Dr = np.dot(dpos, np.linalg.inv(cell))
            D = np.dot(Dr - np.round(Dr) * pbc, cell)
            D_len = np.sqrt((D**2).sum(1))

            cutoff = min(max(D_len), max(diags) / 2.)

    latt_len = np.sqrt((cell**2).sum(1))
    V = abs(np.linalg.det(cell))
    padding = pbc * np.array(np.ceil(cutoff * np.prod(latt_len) /
                                     (V * latt_len)), dtype=int)

    offsets = np.mgrid[
        -padding[0]:padding[0] + 1,
        -padding[1]:padding[1] + 1,
        -padding[2]:padding[2] + 1].T
    tvecs = np.dot(offsets, cell)
    coords = pos[None, None, None, :, :] + tvecs[:, :, :, None, :]

    ncell = np.prod(offsets.shape[:-1])
    index = np.arange(len(atoms))[None, :].repeat(ncell, axis=0).flatten()
    coords = coords.reshape(np.prod(coords.shape[:-1]), 3)
    offsets = offsets.reshape(ncell, 3)

    return index, coords, offsets


def get_area_neighbors(atoms, cell_cutoff=15, cutoff=None,
                       return_std_con=False):
    """
    Extention of the CatKit Voronoi connectivity.

    Return the connectivity matrix from the Voronoi
    method weighted by the area of the Voronoi vertices,
    which changes smoothly with small geometry changes.
    Multi-bonding occurs through periodic boundary conditions.

    Parameters
    ----------
    atoms : atoms object
        Atoms object with the periodic boundary conditions and
        unit cell information to use.
    cell_cutoff : float
        Cutoff for supercell repetition used for Voronoi
    cutoff : float
        Maximum atomic bond distance to consider for connectivity

    Returns
    -------
    connectivity : ndarray (n, n) of floats
       Weighted number of edges formed between atoms in a system.
    connectivity_int : ndarray (n, n) of ints
       Number of edges formed between atoms in a system.
    """
    index, coords, offsets = expand_cell(atoms,
                                         cutoff=cell_cutoff)
    L = int(len(offsets) / 2)

    origional_indices = np.arange(L * len(atoms), (L + 1) * len(atoms))

    areas = []
    area_indices = []
    connectivity = np.zeros((len(atoms), len(atoms)))
    connectivity_int = np.zeros((len(atoms), len(atoms)))
    try:
        voronoi = scipy.spatial.Voronoi(coords, qhull_options='QbB Qc Qs')
        points = voronoi.ridge_points
    except:
        print('Voronoi failed')
        if not return_std_con:
            return connectivity
        else:
            return connectivity, connectivity_int.astype(int)

    Voro_cell = voronoi.points.ptp(axis=0)
    for i, n in enumerate(origional_indices):
        ridge_indices = np.where(points == n)[0]
        p = points[ridge_indices]
        dist = np.linalg.norm(np.diff(coords[p], axis=1), axis=-1)[:, 0]
        edges = np.sort(index[p])

        if not edges.size:
            warnings.warn(
                ("scipy.spatial.Voronoi returned an atom which has "
                 "no neighbors. This may result in incorrect connectivity."))
            continue

        unique_edge = np.unique(edges, axis=0)

        for j, edge in enumerate(unique_edge):
            indices = np.where(np.all(edge == edges, axis=1))[0]
            if cutoff:
                indices = indices[np.where(dist[indices] < cutoff)[0]]
            d = dist[indices]
            count = len(d)
            u, v = edge
            area_indices += [sorted([u, v])]
            edge_areas = []
            for ii, ind in enumerate(ridge_indices[indices]):
                vertices = np.array([list(voronoi.vertices[v]) for v in
                                     voronoi.ridge_vertices[ind]])
                if len(vertices) < 3:
                    continue

                vertices *= Voro_cell

                area = get_weighted_area(vertices, d[ii])
                if area is None:
                    continue

                connectivity[u][v] += area
                connectivity[v][u] += area

            connectivity_int[u][v] += count
            connectivity_int[v][u] += count

    connectivity /= 2

    if not return_std_con:
        return connectivity
    else:
        connectivity_int /= 2
        return connectivity, connectivity_int.astype(int)


def get_voronoi_neighbors(atoms, cutoff=10, return_distances=False):
    """
    Based on connectivity from CatKit, with a fix for the distance
    cutoff.
    Return the connectivity matrix from the Voronoi
    method. Multi-bonding occurs through periodic boundary conditions.

    Parameters
    ----------
    atoms : atoms object
        Atoms object with the periodic boundary conditions and
        unit cell information to use.
    cutoff : float
        Radius of maximum atomic bond distance to consider.

    Returns
    -------
    connectivity : ndarray (n, n)
        Number of edges formed between atoms in a system.
    """
    index, coords, offsets = expand_cell(atoms, cutoff=cutoff)

    xm, ym, zm = np.max(coords, axis=0) - np.min(coords, axis=0)

    L = int(len(offsets) / 2)
    origional_indices = np.arange(L * len(atoms), (L + 1) * len(atoms))

    voronoi = scipy.spatial.Voronoi(coords, qhull_options='QbB Qc Qs')
    points = voronoi.ridge_points

    connectivity = np.zeros((len(atoms), len(atoms)))
    distances = []
    distance_indices = []
    for i, n in enumerate(origional_indices):
        ridge_indices = np.where(points == n)[0]
        p = points[ridge_indices]
        dist = np.linalg.norm(np.diff(coords[p], axis=1), axis=-1)[:, 0]
        edges = np.sort(index[p])

        if not edges.size:
            warnings.warn(
                ("scipy.spatial.Voronoi returned an atom which has "
                 "no neighbors. This may result in incorrect connectivity."))
            continue

        unique_edge = np.unique(edges, axis=0)

        for j, edge in enumerate(unique_edge):
            indices = np.where(np.all(edge == edges, axis=1))[0]
            d = dist[indices][np.where(dist[indices] < cutoff)[0]]
            count = len(d)
            if count == 0:
                continue

            u, v = edge

            distance_indices += [sorted([u, v])]
            distances += [sorted(d)]

            connectivity[u][v] += count
            connectivity[v][u] += count

    connectivity /= 2
    if not return_distances:
        return connectivity.astype(int)

    if len(distances) > 0:
        distance_indices, unique_idx_idx = \
            np.unique(distance_indices, axis=0, return_index=True)
        distance_indices = distance_indices.tolist()

        distances = [distances[i] for i in unique_idx_idx]

    pair_distances = {'indices': distance_indices,
                      'distances': distances}

    return connectivity.astype(int), pair_distances


def get_weighted_area(vertices, d):
    """ helper method for voronoi area"""
    center = vertices[0, :]
    vertices -= center

    # Plane of polygon in 3d defined by unit vectors ux and uy
    ux = vertices[1, :].copy()
    ux /= np.linalg.norm(ux)
    vy = vertices[2, :].copy()

    # ensure uy is orthogonal to ux
    vyproj = np.dot(vy, ux) * ux
    uy = vy - vyproj
    uy /= np.linalg.norm(uy)

    # Project all vectors onto plane to get 2D coordinates
    Xs = np.dot(vertices, ux)
    Ys = np.dot(vertices, uy)

    poly_coords = [[Xs[i], Ys[i]] for i in range(len(Xs))]
    polygon = Polygon(poly_coords)

    try:
        area = polygon.area
    except:
        return None

    perimeter = polygon.length

    # Scale area with bond distance and perimeter
    weight = d * perimeter / 4
    area /= weight

    return area


def get_covalent_density(atoms):

    covalent_radii = np.array([cradii[n] for n in atoms.numbers])
    covalent_volume = np.sum(4/3 * np.pi * covalent_radii ** 3)
    cell_volume = atoms.get_volume()
    density = covalent_volume / cell_volume

    return density


def get_apf_fitness(atoms, apf_0=0.1):
    """ Volume dependent fitness function, expressed in terms of the 
    atomic packing factor V_atoms/V_cell assuming a parabolic dependency
    F ~ 1/V**2

    Parameters:
    apf_0:  APF value where fitness=0
    """

    apf = get_covalent_density(atoms)
    a0 = apf_0**2/(1 - apf_0**2)

    fitness = 1 + a0 * (1 - 1 / apf**2)

    return fitness


def get_fitness(atoms):
    N_metal = len([a for a in atoms if a.number in metal_numbers])
    symbols = atoms.get_chemical_symbols()

    if N_metal == len(atoms):
        # If metal use volume as fitness
        fitness = - atoms.get_volume()
    else:
        fitness = - get_ewald_energy(atoms)

    return fitness


def get_oxidation_states(atoms):

    metal_idx = [i for i, a in enumerate(
        atoms) if a.number in metal_numbers]

    non_metal_idx = [i for i in range(len(atoms)) if not i in metal_idx]

    symbols = atoms.get_chemical_symbols()

    fix_oxi_nonM = np.array([fixed_oxi_states[symbols[i]]
                             for i in non_metal_idx])

    if len(fix_oxi_nonM) == 0:
        raise NotImplementedError('Fitness for {} composition not implemented'
                                  .format(atoms.get_chemical_formula))

    oxi_states = np.array([-2] * len(atoms), dtype=float)
    oxi_states[non_metal_idx] = fix_oxi_nonM
    con_matrix = get_area_neighbors(atoms)
    for mi in metal_idx:
        M_nonM_connectivity = con_matrix[mi][non_metal_idx]

        norm = np.sum(con_matrix[non_metal_idx][:, metal_idx], axis=-1)

        idx = np.where(norm > 0)[0]

        oxi_states[mi] = -sum(M_nonM_connectivity[idx]
                              * fix_oxi_nonM[idx] / norm[idx])
    return oxi_states


def get_ewald_energy(atoms, use_density=True):
    oxi_states = get_oxidation_states(atoms)

    structure = AseAtomsAdaptor.get_structure(atoms)
    structure.add_oxidation_state_by_site(oxi_states)
    e = EwaldSummation(structure).total_energy

    return e / len(atoms)


def get_oxy_fitness(atoms):
    a = 1


def get_optimal_oxidation_states_for_composition(metal_symbols, n_O):
    """
    A_nB_mO_k with oxidation states O_A and O_B setting O_O = 2
    Solve for integers O_A and O_B
    n_A * O_A + n_B * O_B = n_O * 2
    """

    n_M = len(metal_symbols)
    avg_oxi_state = n_O / n_M * 2

    oxi_states_dict = {}
    metal_symbols, counts = np.unique(metal_symbols, return_counts=True)

    electroneg = [electronegs.get(m, 0) for m in metal_symbols]
    indices = np.argsort(electroneg)[::-1]
    metal_symbols = metal_symbols[indices]

    #n_M_list = [list(metal_symbols).count(m) for m in metal_symbols]
    # print(n_M_list)
    pref_O_states = [prefered_O_state[m] for m in metal_symbols]

    if len(metal_symbols) == 2:
        # Make a guess for favorable oxidation states
        oxy_matches = []
        n_A, n_B = counts
        for o_A in range(1, n_O * 2 // n_A):
            o_B = (2 * n_O - n_A * o_A) / n_B
            if o_B % 1 == 0:
                oxy_matches += [[o_A, int(o_B)]]

        oxy_matches = np.array(oxy_matches)

        pref_O_M = np.repeat(np.expand_dims(pref_O_states, 0),
                             len(oxy_matches[:, 0]), 0)

        Oxy_fitness = np.sum(np.abs(oxy_matches - pref_O_M), axis=1)

        best_fit = np.argmin(Oxy_fitness)

        oxy_state_list = oxy_matches[best_fit]

    elif len(metal_symbols) == 1:  # only one type of metal
        oxy_state_list = [avg_oxi_state]

    oxi_states = {}
    for i, m in enumerate(metal_symbols):
        oxi_states.update({m: int(oxy_state_list[i])})

    return oxi_states


def get_connections(atoms, decimals=1):

    connectivity = get_area_neighbors(atoms)

    atoms_connections = {}

    for i in range(len(atoms)):
        for j in range(len(atoms)):
            symbols = '-'.join(sorted([atoms[i].symbol, atoms[j].symbol]))
            if not symbols in atoms_connections:
                atoms_connections[symbols] = [0 for n in range(len(atoms))]

            if not atoms[i].symbol == atoms[j].symbol:
                idx = np.argsort([atoms[i].symbol, atoms[j].symbol])
                k = [i, j][idx[-1]]
            else:
                k = i

            if not connectivity[i][j]:
                continue
            atoms_connections[symbols][k] += \
                connectivity[i][j] / 2

    final_connections = {}
    for key, value in atoms_connections.items():
        value = [v for v in value if v > 0]
        if len(value) > 0:
            m = np.mean(value)
            std = np.std(value)
        else:
            m = 0
            std = 0
        final_connections[key + '_mean'] = np.round(m, decimals)
        final_connections[key + '_std'] = np.round(std, decimals)
    return final_connections



import sys
import numpy as np
from spglib import get_symmetry_dataset, standardize_cell
from ase import Atoms


class SpglibInterface:
    """Prototype classification of atomic structure in ASE Atoms format"""

    def __init__(self, atoms, tolerance=1e-3):

        self.atoms = atoms

        self.spglibdata = get_symmetry_dataset((atoms.get_cell(),
                                                atoms.get_scaled_positions(),
                                                atoms.get_atomic_numbers()),
                                               symprec=tolerance)
        self.tolerance = tolerance
        self.spacegroup = self.spglibdata['number']

    def get_spacegroup(self):
        return self.spglibdata['number']

    def get_conventional_atoms(self):
        """Transform from primitive to conventional cell"""

        std_cell = self.spglibdata['std_lattice']
        positions = self.spglibdata['std_positions']
        numbers = self.spglibdata['std_types']

        atoms = Atoms(numbers=numbers,
                      cell=std_cell,
                      pbc=True)

        atoms.set_scaled_positions(positions)

        atoms.wrap()

        return atoms

    def get_primitive_atoms(self, atoms):
        """Transform to primitive cell"""

        lattice, scaled_positions, numbers = standardize_cell(atoms,
                                                              to_primitive=True,
                                                              no_idealize=True,
                                                              symprec=1e-5)

        atoms = Atoms(numbers=numbers,
                      cell=lattice,
                      pbc=True)

        atoms.set_scaled_positions(scaled_positions)

        atoms.wrap()

        return atoms



from ase.calculators.general import Calculator
from ase.calculators.calculator import PropertyNotImplementedError


class DummyCalc(Calculator):
    name = 'DummyCalc'
    implemented_properties = ['energy']

    def __init__(self,
                 energy_zero=None,
                 track_output=False,
                 ):
        self.energy_zero = energy_zero
        self.atoms = None

    def todict(self):
        out_dict = {"key0": 0, "key1": 1}
        return(out_dict)

    def get_potential_energy(self, force_consistent=False):
        return self.energy_zero

    def check_state(self, atoms, tol=1e-15):
        """Check for system changes since last calculation."""
        from ase.calculators.calculator import all_changes, equal
        if self.atoms is None:
            system_changes = all_changes[:]
        else:
            system_changes = []
            if not equal(self.atoms.positions, atoms.positions, tol):
                system_changes.append('positions')
            if not equal(self.atoms.numbers, atoms.numbers):
                system_changes.append('numbers')
            if not equal(self.atoms.cell, atoms.cell, tol):
                system_changes.append('cell')
            if not equal(self.atoms.pbc, atoms.pbc):
                system_changes.append('pbc')
            if not equal(self.atoms.get_initial_magnetic_moments(), atoms.get_initial_magnetic_moments(), tol):
                system_changes.append('initial_magmoms')
            if not equal(self.atoms.get_initial_charges(), atoms.get_initial_charges(), tol):
                system_changes.append('initial_charges')

        return(system_changes)

    def get_property(self, name, atoms=None, allow_calculation=True):
        """Returns the value of a property"""
        if name not in DummyCalc.implemented_properties:
            raise PropertyNotImplementedError
        if atoms is None:
            atoms = self.atoms

        saved_property = {'energy': 'energy_zero'}
        if hasattr(self, saved_property[name]):
            result = getattr(self, saved_property[name])
        else:
            result = None

        return(result)



import numpy as np
import pylab as plt
import string

from pprint import pprint

from protosearch import build_bulk

path = build_bulk.__path__[0]

wyckoff_data = path + '/Wyckoff.dat'
wyckoff_pairs = path + '/Symmetric_Wyckoff_Pairs.dat'


class BulkEnumerator:

    def __init__(
            stoichiometry='',
            spacegroup_start=1,
            spacegroup_end=230):

        self.stoichiometry = stoichiometry

        self.spacegroups = range(spacegroup_start, spacegroup_end + 1)


def get_wyckoff_combinations_for_spacegroup(spacegroup, stoichiometry,
                                            n_max_atoms=6):

    letters, multiplicity = get_wyckoff_letters_and_multiplicity(spacegroup)

    pair_names, M, free_letters = get_wyckoff_pair_symmetry_matrix(spacegroup)
    n_pairs = len(pair_names)
    #print(pair_names[:17])
    #print(M[:17][:, :17])
    #sys.exit()
    stoich_vector = stoichiometry.split('_')
    n_types = len(stoich_vector)
    n_atoms = [int(n) for n in stoich_vector]
    n_tot = sum(n_atoms)

    n_rep = n_max_atoms // n_tot
    n_atoms_rep = [n * n_rep for n in n_atoms]

    start_index = 0

    combinations = [['']]

    letter_combinations = [[]]

    pair_indices = []

    final_combinations = []

    n = 0

    natoms = []

    natoms_combinations = []

    while len(pair_combinations) > 0:
        prev_combinations = pair_combinations.copy()
        combinations = []
        prev_natoms_combinations = natoms_combinations.copy()
        natoms_combinations = []
        #prev_letter_combinations = prev_letter_combinations.copy()
        for j, prev_comb in enumerate(prev_combinations):
            next_letter = prev_comb[-1].split('_')[-1]
            if next_letter == '':
                next_pair_indices = range(len(pair_names))
                start_index = 0
            else:
                next_pair_indices = [i for i, p in enumerate(
                    pair_names) if p[0] == next_letter]
                start_index = next_pair_indices[0] - 1

            excl_indices = []
            for i in next_pair_indices:
                pair = pair_names[i]

                check_range = np.array(
                    [k for k in range(start_index, i) if not k in excl_indices])

                if len(check_range) == 0:
                    check_range = []
                    continue
                #print(pair, check_range)
                #print(M[i, check_range])
                # np.all(M[check_range, first_match] == 0):
                if np.all(M[i][check_range] == 0):
                    # print(pair)

                    if prev_comb == ['']:
                        new_comb = [pair]
                    else:
                        new_comb = prev_comb + [pair]

                    last_letter = new_comb[-1][-1]
                    comb_letters = [c[0] for c in new_comb] + [last_letter]

                    if not last_letter in free_letters and last_letter in comb_letters[:-1]:
                        #print(comb_letters, last_letter)
                        #start_index = i + 1
                        excl_indices += [i]
                        #prev_indices += [i]
                        # print(new_comb)
                        # Only add position once
                        continue

                    comb_natoms = [multiplicity[np.where(
                        letters == l)[0]] for l in comb_letters]

                    #cumsum_
                    if sum(comb_natoms) > sum(n_atoms_rep):
                        continue
                    

                    if sum(comb_natoms) % n_tot == 0:

                        rep = sum(comb_natoms) / n_tot
                        n_atoms_temp = [n * rep for n in n_atoms]

                        last_i = 0
                        match = True
                        atoms_wyckoff_cut = []
                        for na in n_atoms_temp:
                            cumsum_atoms = np.cumsum(comb_natoms[last_i:])
                            idx = np.where(cumsum_atoms==na)[0]
                            if len(idx) == 0:
                                match = False
                            else:
                                atoms_wyckoff_cut += [idx[0] + 1]
                                last_i = idx[0] + 1
                        #print(atoms_wyckoff_cut)
                        #if not idx[0] + 2 == len(comb_natoms) or not match:
                        #    continue

                        if not match:
                            continue
                        
                        print(new_comb)
                        unique = True
                        pair_test = new_comb.copy()
                        if len(new_comb) > 1:
                            pair_test += [new_comb[0]
                                          [0] + '_' + new_comb[1][-1]]

                        indices_1 = [pair_names.index(p) for p in pair_test]
                        for f_comb in [f for f in final_combinations if len(f) == len(new_comb)]:
                            print(new_comb, f_comb)
                            pair_test_2 = f_comb.copy()
                            if len(f_comb) > 1:
                                pair_test_2 += [f_comb[0]
                                                [0] + '_' + f_comb[1][-1]]

                            indices_0 = [pair_names.index(
                                p) for p in pair_test_2]

                            cut0 = 0
                            uniques = []
                            for cut in atoms_wyckoff_cut:
                                idx0 = indices_0[cut0: cut]
                                idx1 = indices_1[cut0: cut]
                                cut0 = cut.copy()
                                M_test = M[idx0, :][:, idx1]
                                if np.all(M_test.any(axis=0)) and np.all(M_test.any(axis=1)):
                                    print(idx0, idx1, M_test)
                                    print('no')
                                    uniques += [False]
                                else:
                                    uniques += [True]
                            unique += np.any(uniques)
                        #print(unique)
                        if unique:
                            final_combinations += [new_comb]
                            #print(new_comb, 'All ok')
                    combinations += [new_comb]
                    print(combinations)
            # print(final_combinations)
            # sys.exit()
        n += 1

    final_names = []
    for final_comb in final_combinations:
        last_letter = final_comb[-1][-1]
        comb_letters = [c[0] for c in final_comb] + [last_letter]
        comb_multi = [multiplicity[np.where(letters == l)[0]]
                      for l in comb_letters]

        n_rep = sum(comb_multi) / n_tot
        n_atoms_rep = [n * n_rep for n in n_atoms]

        name = ''
        atom_i = 0

        atoms_count = 0
        for i, l in enumerate(comb_letters):
            if atoms_count == n_atoms_rep[atom_i]:
                atom_i += 1
                atoms_count = 0
                name += '_'
            if atoms_count < n_atoms_rep[atom_i]:
                name += l
                atoms_count += comb_multi[i]
            elif atoms_count > n_atoms_rep[atom_i]:
                name = None

        if name:
            name = '_'.join([''.join(sorted(list(n)))
                             for n in name.split('_')])
            final_names += [name]

    final_names = sorted(list(set(final_names)))
    print(len(final_names))
    pprint(final_names)


def get_wyckoff_letters_and_multiplicity(spacegroup):
    letters = np.array([], dtype=str)
    multiplicity = np.array([])
    with open(wyckoff_data, 'r') as f:
        i_sg = np.inf
        i_w = np.inf
        for i, line in enumerate(f):
            if '1 {} '.format(spacegroup) in line \
               and not i_sg < np.inf:
                i_sg = i
            if i > i_sg:
                if len(line) > 15:
                    continue
                if len(line) == 1:
                    break
                multi, w, sym, sym_multi = line.split(' ')
                letters = np.insert(letters, 0, w)
                multiplicity = np.insert(multiplicity, 0, multi)

    return letters, multiplicity


def get_wyckoff_pair_symmetry_matrix(spacegroup):
    letters, multiplicity = get_wyckoff_letters_and_multiplicity(spacegroup)

    n_points = len(letters)
    pair_names = []
    for i in range(n_points):
        for j in range(n_points):
            pair_names.append('{}_{}'.format(letters[i], letters[j]))

    M = np.zeros([n_points**2, n_points**2])

    with open(wyckoff_pairs, 'r') as f:
        sg = 1
        for i, line in enumerate(f):
            if len(line) == 1:
                if sg < spacegroup:
                    sg += 1
                    continue
                else:
                    break
            if sg < spacegroup:
                continue
            w_1 = line[:3]
            if not w_1 in pair_names:
                continue
            k = pair_names.index(w_1)
            pairs0 = line.split('\t')[2: -1]
            for w_2 in pairs0:
                j = pair_names.index(w_2)
                M[k, j] = 1

    free_letters = []
    for l in letters:
        i = pair_names.index(l + '_' + l)
        if M[i, i] == 1:
            free_letters += [l]

    np.fill_diagonal(M, 1)

    return pair_names, M, free_letters


if __name__ == "__main__":
    #pair_names, M = get_wyckoff_pair_symmetry_matrix(3)
    get_wyckoff_combinations_for_spacegroup(10, '1_2', n_max_atoms=3)



from .workflow import Workflow, MLinterface


class Acqusition(Workflow, MLInterface):
    """ TODO: Decide which calculations to do next. 
    Should interface MLInterface for predictions.
    """

    def __init__(self):
        super().__init__()

    def get_prediction(self):
        prototypes = None
        return prototypes

    def update_model(self):
        """ Should we call the ML model here, 
        to update with completed calculations? """

        self.collect()  # collect completed calculations
        return None

    def submit_batch(self, prototypes, species):
        """Call WorkFlow module to submit selected prototypes"""

        prototypes = self.get_prediction()
        for prototype in prototypes:
            self.submit(prototype)



import pylab as p
import numpy as np
from protosearch.utils.standards import VaspStandards

from .workflow import Workflow, PrototypeSQL


class Convergence(Workflow):
    """TO DO: Automated convergence check """

    def __init__(self,
                 calculator='vasp',
                 basepath_ext=None):

        super().__init__(calculator=calculator,
                         basepath_ext=basepath_ext)
        self._connect()

        self.fixed_convergence_params = ['encut', 'kspacing', 'ismear']

        self.fix_params = {}
        for p in self.fixed_convergence_params:
            self.fix_params.update({p: VaspStandards.calc_parameters[p]})
        # index for highest converged point in numerically sorted array
        self.convergence_index = {'encut': -1, 'kspacing': 0, 'ismear': 0}

    def get_parametrized_energies(self, check_parameters=['encut', 'kspacing']):

        total_energies = {}
        for param in check_parameters:
            param_st = VaspStandards.calc_parameters[param]
            prototypes = []
            for d in self.ase_db.select('{}!={}'.format(param, param_st)):
                if not d.p_name + '|' + d.formula in prototypes:
                    prototypes.append(d.p_name + '|' + d.formula)

            total_energies[param] = {}
            for proto in prototypes:
                total_energies[param][proto] = {}
                name, formula = proto.split('|')
                fix_params = self.fix_params.copy()
                del fix_params[param]
                for d in self.ase_db.select(p_name=name,
                                            relaxed=1,
                                            formula=formula,
                                            **fix_params):

                    value = d.get(param, None)
                    total_energies[param][proto][value] = d.energy

        return total_energies

    def plot_convergence(self, check_parameters=['encut', 'kspacing']):

        total_energies = self.get_parametrized_energies()

        for param, proto_dict in total_energies.items():
            p.figure()
            p.title(param)
            for proto, points in proto_dict.items():
                x = np.array(list(points.keys()))
                y = np.array(list(points.values()))

                if len(x) <= 1:
                    continue

                index = np.argsort(x)
                x = x[index]
                y = y[index]

                y -= y[self.convergence_index[param]]
                p.plot(x, np.zeros_like(x), 'r--')

                p.plot(x, y, '-*', label=proto)
                p.ylim([-0.05, 0.05])
                p.legend()
        p.show()

    def submit_convergence_check(self,
                                 prototype,
                                 check_parameters=['encut',
                                                   'kspacing',
                                                   'sigma']):

        parameter_mapping = {'encut': [300, 400, 500, 525, 550,
                                       600, 700, 800],
                             'kspacing': [10, 15, 20, 25, 30, 40],
                             'sigma': [0.05, 0.1, 0.15, 0.2]}

        for param in check_parameters:
            for value in parameter_mapping[param]:
                self.submit(prototype, calc_parameters={param: value})



import os
import json
import subprocess
import time
import numpy as np
import copy
import ase
import ase.build
from ase.io import read

from protosearch.build_bulk.build_bulk import BuildBulk
from protosearch.build_bulk.classification import PrototypeClassification

from protosearch.utils.standards import VaspStandards, CrystalStandards
from protosearch.calculate.dummy_calc import DummyCalc
from protosearch.calculate.submit import get_submitter
from .prototype_db import PrototypeSQL

standard_lattice = CrystalStandards.standard_lattice_mp
all_elements = list(standard_lattice.keys()) + ['H', 'N', 'O']


class Workflow(PrototypeSQL):
    """Submits calculations on specified cluster, and tracks the calculations
    in an ASE db.
    Parameters:

    db_filename: str
       full path to db file to save calculations
    cluster: str
       Currently only 'tri' is working
    """
    def __init__(self,
                 db_filename,
                 calculator='vasp',
                 cluster='tri',
                 verbose=False):

        self.verbose = verbose

        super().__init__(filename=db_filename)

        self.Submitter = get_submitter(cluster)

        self.collected = False

    def _collect(self):
        if self.collected:
            return
        self.recollect()

    def recollect(self):
        for row in self.ase_db.select(completed=0, submitted=1):
            subprocess.call('trisync', cwd=row.path)

        self.collected = True

    def submit_atoms_batch(self, atoms_list, ncpus=None, calc_parameters=None,
                           **kwargs):
        """Submit a batch of calculations. Takes a list of atoms
        objects as input"""
        batch_no = self.get_next_batch_no()
        for atoms in atoms_list:
            self.submit_atoms(atoms, ncpus, batch_no,
                              calc_parameters, **kwargs)

    def submit_atoms(self, atoms, ncpus=None, batch_no=None,
                     calc_parameters=None, **kwargs):
        """Submit a calculation for an atoms object"""
        PC = PrototypeClassification(atoms)
        prototype, cell_parameters = PC.get_classification()

        if self.is_calculated(formula=atoms.get_chemical_formula(),
                              p_name=prototype['p_name']):
            return

        Sub = self.Submitter(atoms=atoms,
                             ncpus=ncpus,
                             calc_parameters=calc_parameters)

        Sub.submit_calculation()

        key_value_pairs = {'p_name': prototype['p_name'],
                           'path': Sub.excpath,
                           'spacegroup': prototype['spacegroup'],
                           'wyckoffs': json.dumps(prototype['wyckoffs']),
                           'species': json.dumps(prototype['species']),
                           'ncpus': Sub.ncpus}

        key_value_pairs.update(kwargs)

        if batch_no:
            key_value_pairs.update({'batch': batch_no})

        self.write_submission(key_value_pairs)

    def submit_batch(self, prototypes, ncpus=None,
                     calc_parameters=None, **kwargs):
        """Submit a batch of calculations. Takes a list of prototype
        dicts as input"""
        batch_no = self.get_next_batch_no()
        for prototype in prototypes:
            self.submit(prototype, ncpus, batch_no, calc_parameters, **kwargs)
        self.write_status(last_batch_no=batch_no)

    def submit(self, prototype, ncpus=None, batch_no=None, calc_parameters=None,
               **kwargs):
        """Submit a calculation for a prototype, generating atoms
        with build_bulk and enumerator"""
        cell_parameters = prototype.get('parameters', None)

        BB = BuildBulk(prototype['spacegroup'],
                       prototype['wyckoffs'],
                       prototype['species'],
                       )

        atoms = BB.get_atoms(cell_parameters=cell_parameters)
        p_name = BB.get_prototype_name(prototype['species'])
        formula = atoms.get_chemical_formula()

        Sub = self.Submitter(atoms=atoms,
                             ncpus=ncpus,
                             calc_parameters=calc_parameters)

        Sub.submit_calculation()

        key_value_pairs = {'p_name': p_name,
                           'path': Sub.excpath,
                           'spacegroup': BB.spacegroup,
                           'wyckoffs': json.dumps(BB.wyckoffs),
                           'species': json.dumps(BB.species),
                           'ncpus': Sub.ncpus}

        key_value_pairs.update(kwargs)

        if batch_no:
            key_value_pairs.update({'batch': batch_no})

        self.write_submission(key_value_pairs)

    def submit_enumerated(self, map_species, selection={}):
        """
        Submit a collection of enumeated prototypes.
        Parameters:
        ----------
        map_species: dict
          map A0, A1, etc to species. F.ex: {'A0': 'Fe', 'A1': 'O'}
        selection: dict
          ase db selection. F.ex: {'spacegroup': 166, natom=3}
        """

        prototypes = self.select(**selection)

        for prototype in prototypes:
            self.submit(prototype)

    def submit_id_batch(self, calc_ids, ncpus=None,
                        calc_parameters=None, **kwargs):
        """Submit a batch of calculations. Takes a list of atoms
        objects db ids as input"""
        batch_no = self.get_next_batch_no()
        for calc_id in calc_ids:
            self.submit_id(calc_id, ncpus, batch_no, calc_parameters)
        self.write_status(last_batch_no=batch_no)

    def submit_id(self, calc_id, ncpus=None, batch_no=None,
                  calc_parameters=None, **kwargs):
        """
        Submit an atomic structure by id
        """
        row = self.ase_db.get(id=int(calc_id))
        atoms = row.toatoms()

        formula = row.formula
        p_name = row.p_name

        if self.is_calculated(formula=formula,
                              p_name=p_name):
            return

        Sub = self.Submitter(atoms=atoms,
                             ncpus=ncpus,
                             calc_parameters=calc_parameters)
        Sub.submit_calculation()

        key_value_pairs = {'path': Sub.excpath,
                           'submitted': 1,
                           'ncpus': Sub.ncpus}

        if batch_no is not None:
            key_value_pairs.update({'batch': batch_no})

        key_value_pairs.update(kwargs)

        self.ase_db.update(int(calc_id), **key_value_pairs)

    def write_submission(self, key_value_pairs):
        """Track submitted job in database"""
        con = self.connection or self._connect()
        self._initialize(con)

        atoms = read(key_value_pairs['path'] + '/initial.POSCAR')

        key_value_pairs.update({'relaxed': 0,
                                'completed': 0,
                                'submitted': 1})

        self.ase_db.write(atoms, key_value_pairs)

    def check_submissions(self, **kwargs):
        """Check for completed jobs"""
        if not self.collected:
            self._collect()
        con = self.connection or self._connect()
        self._initialize(con)

        status_count = {'completed': 0,
                        'running': 0,
                        'errored': 0}
        completed_ids = []
        failed_ids = []
        running_ids = []
        for d in self.ase_db.select(submitted=1, completed=0, **kwargs):
            path = d.path + '/simulation'
            calcid0 = d.id
            status, calcid = self.check_job_status(path, calcid0)
            status_count[status] += 1
            if status == 'completed':
                completed_ids += [calcid]
            elif status == 'errored':
                failed_ids += [calcid]
            elif status == 'running':
                running_ids += [calcid]

        if self.verbose:
            print('Status for calculations:')
            for status, value in status_count.items():
                print('  {} {}'.format(value, status))
        return completed_ids, failed_ids, running_ids

    def check_job_status(self, path, calcid):
        status = 'running'
        for root, dirs, files in os.walk(path):
            if 'monitor.sh' in files and not 'finalized' in files:
                status = 'running'
                self.ase_db.update(id=calcid, runpath=root)
                break
            elif 'monitor.sh' in files and 'completed' in files:
                # Calculation completed - now save everything
                try:  # Sometimes outcar is corrupted
                    atoms = ase.io.read(root + '/OUTCAR')
                    status = 'completed'
                    calcid = self.save_completed_calculation(atoms,
                                                             root,
                                                             calcid)
                except BaseException as e:
                    print("Couldn't read OUTCAR", calcid)
                    status = 'errored'
                    self.save_failed_calculation(root, calcid)
                break
            elif 'monitor.sh' in files and 'finalized' in files:
                status = 'errored'
                self.save_failed_calculation(root, calcid)
                break

        return status, calcid

    def save_completed_calculation(self, atoms, runpath, calcid,
                                   read_params=True):

        batch_no = self.ase_db.get(id=calcid).get('batch', None)

        PC = PrototypeClassification(atoms)
        prototype, cell_parameters = PC.get_classification()

        key_value_pairs = self.ase_db.get(id=calcid).get('key_value_pairs', {})

        key_value_pairs.update({'relaxed': 1,
                                'completed': 1,
                                'initial_id': calcid,
                                'runpath': runpath})
        key_value_pairs.update(prototype)
        key_value_pairs.update(cell_parameters)

        if batch_no:
            key_value_pairs.update({'batch': batch_no})

        param_dict = {}
        if read_params:
            param_dict = params2dict(runpath + '/param')
            key_value_pairs.update(param_dict)

        atoms = set_calculator_info(atoms, param_dict)

        key_value_pairs = clean_key_value_pairs(key_value_pairs)

        newcalcid = self.ase_db.write(atoms, key_value_pairs)
        self.ase_db.update(id=calcid,
                           final_id=newcalcid,
                           completed=1)

        return newcalcid

    def save_failed_calculation(self, runpath, calcid):

        key_value_pairs = {'completed': -1,
                           'runpath': runpath}
        param_dict = params2dict(runpath + '/param')
        atoms = ase.io.read(runpath + '/initial.POSCAR')

        PC = PrototypeClassification(atoms)
        prototype, cell_parameters = PC.get_classification()

        key_value_pairs.update(prototype)
        key_value_pairs.update(cell_parameters)
        key_value_pairs.update(param_dict)

        key_value_pairs = clean_key_value_pairs(
            key_value_pairs)
        data = {}
        if os.path.isfile(runpath + '/err'):
            with open(runpath + '/err', 'r') as errorf:
                data.update({'error': errorf.read().replace("'", '')})

        if os.path.isfile(runpath + '/err.relax'):
            with open(runpath + '/err.relax', 'r') as errorf:
                data.update({'error_relax': errorf.read().replace("'", '')})

        self.ase_db.update(id=calcid,
                           **key_value_pairs,
                           data=data)

    def rerun_failed_calculations(self):
        self._collect()
        con = self.connection or self._connect()
        self._initialize(con)

        for d in self.ase_db.select(completed=-1):
            error = d.data.get('error', None)
            failed_relax = False
            if error is None:
                failed_relax = True
                error = d.data.get('error_relax', None)

            if not error:
                continue
            resubmit, fail_reason = vasp_errors(error)
            if failed_relax:
                resubmit = True
            if not resubmit:
                if fail_reason == 'ase read':
                    atoms = ase.io.read(d.runpath + '/OUTCAR')
                    calcid = self.save_completed_calculation(atoms,
                                                             d.runpath,
                                                             d.id)
                else:
                    print('Job not resubmitted: {}'.format(fail_reason))
                continue
            ncpus = None
            if fail_reason == 'ncpus':
                ncpus = d.get('ncpus', 1) * 2
            elif fail_reason == 'ase read':
                atoms = read(d.runpath + '/OUTCAR.relax')
                self.ase_db.update(id=d.id, atoms=atoms)
            p_name = d.p_name
            path = d.path + '/simulation'

            calc_parameters = {}
            for param in VaspStandards.sorted_calc_parameters:
                if d.get(param, None):
                    calc_parameters.update({param: d.get(param)})
            if fail_reason == 'Symmetry error':
                calc_parameters.update({'kgamma': True})
            self.submit_id(d.id,
                           ncpus=ncpus,
                           calc_parameters=calc_parameters)

            rerun = (d.get('rerun') or 0) + 1
            self.ase_db.update(id=d.id, completed=0, rerun=rerun)

    def get_completed_batch(self):
        ids = self.check_submissions()

    def reclassify(self, tolerance=0.5e-3):
        """Reclassify prototypes of all completed structures"""
        self._collect()
        con = self.connection or self._connect()
        self._initialize(con)

        for row in self.ase_db.select(relaxed=1):

            PC = PrototypeClassification(row.toatoms())
            prototype, cell_parameters = PC.get_classification()
            prototype.update(cell_parameters)
            prototype = clean_key_value_pairs(prototype)
            self.ase_db.update(row.id, **prototype)

    def submit_standard_states(self, elements=None, batch_no=None):
        elements = elements or all_elements
        for e in elements:
            if e in ['H', 'N', 'O', 'F', 'Cl']:
                atoms = ase.build.molecule(e + '2')
                atoms.set_cell(10 * np.identity(3))
                atoms.center()
                calc_parameters = VaspStandards.molecule_calc_parameters
                self.submit_atoms(atoms, batch_no=batch_no,
                                  calc_parameters=calc_parameters,
                                  standard_state=1)
            else:
                prototype = standard_lattice[e]
                self.submit(prototype, batch_no=batch_no, standard_state=1)


class DummyWorkflow(Workflow):
    """
    """

    def __init__(self,
                 job_complete_time=0.6,
                 *args, **kwargs,
                 ):
        print("USING DUMMY WORKFLOW CLASS | NO DFT SUBMISSION")

        super().__init__(*args, **kwargs)

        self.job_complete_time = job_complete_time

        # Pretend that instance is always in "collected" state
        self.collected = True

        # Will be dotted with fingerprints to produce dummy 'energy' output
        np.random.seed(0)
        self.random_vect = np.random.rand(1, 1000)[0]

    def recollect(self):
        # print("Not actually calling trisync here")
        return(None)

    def submit_id(self, calc_id, ncpus=None,
                  batch_no=None, calc_parameters=None):
        """
        Submit an atomic structure by id
        """
        row = self.ase_db.get(id=int(calc_id))
        # atoms = row.toatoms()

        formula = row.formula
        p_name = row.p_name

        if self.is_calculated(formula=formula, p_name=p_name):
            return

        key_value_pairs = {
            "path": "TEMP",
            "submit_time": time.time(),
            "submitted": 1,
        }

        if batch_no is not None:
            key_value_pairs.update({'batch': batch_no})

        self.ase_db.update(int(calc_id), **key_value_pairs)

    def check_job_status(self, path, calcid):
        """
        """
        d = self.ase_db.get(id=calcid)

        status = 'running'
        atoms = d.toatoms()
        calcid = d.id

        time_submit = d.submit_time
        time_i = time.time()
        time_elapsed = time_i - time_submit

        if time_elapsed > self.job_complete_time:
            # Job completed
            status = 'completed'

            train_features, train_target = self.get_fingerprints([calcid])
            shape = train_features.shape
            random_vect = self.random_vect[0:shape[-1]]
            dummy_energy = random_vect.dot(train_features[0]) / 1000

            atoms = copy.deepcopy(atoms)
            calc = DummyCalc(energy_zero=dummy_energy)
            atoms.set_calculator(calc)

            path = d.path
            runpath = "TEMP"

            new_calcid = self.save_completed_calculation(
                atoms, path, runpath, calcid,
                read_params=False)
            calcid = new_calcid

        else:
            # Job not completed
            status = 'running'

        return(status, calcid)


def clean_key_value_pairs(key_value_pairs):
    for key, value in key_value_pairs.items():
        if isinstance(value, list):
            key_value_pairs[key] = json.dumps(value)
        try:
            value = int(value)
            key_value_pairs[key] = value
        except:
            pass
    return key_value_pairs


def set_calculator_info(atoms, parameters):

    atoms.calc.name = 'vasp.5.4.4.18Apr17-6-g9f103f2a35'
    atoms.calc.parameters = parameters

    return atoms


def vasp_errors(error):

    if 'lattice_constant = float(f.readline().split()[0])' in error:
        return False, 'Vasp failed'
    elif 'Vasp exited with exit code:' in error:
        return True, 'ncpus'
    elif "local variable 'forces' referenced before assignment" in error:
        return True, 'Symmetry error'
    elif 'could not convert string to float' in error:
        return True, 'ase read'
    elif 'local variable forces referenced before assignment' in error:
        return True, 'ase read'
    elif 'invalid literal for float()' in error:
        return True, 'ase read'
    else:
        return False, ''


def params2dict(paramfile):
    param_dict = {}
    with open(paramfile, 'r') as f:
        param = f.read().lstrip('/').rstrip('\n')
        param_values = param.split('/')
        for i, param_key in enumerate(VaspStandards.sorted_calc_parameters):
            param_dict[param_key] = param_values[i]

    return param_dict



import numpy as np
from numpy.random import rand, randint, shuffle
from ase.visualize import view
from ase.geometry import get_distances, cell_to_cellpar
from ase.data import atomic_numbers as a_n
from ase.data import covalent_radii as cradii
from ase.spacegroup import crystal

from protosearch.ml_modelling.regression_model import get_regression_model
from protosearch.ml_modelling.fingerprint import clean_features
from .wyckoff_symmetries import WyckoffSymmetries
from .fitness_function import get_fitness, get_connections


class CellParameters(WyckoffSymmetries):
    """
    Provides a fair estimate of cell parameters including lattice constants,
    angles, and free wyckoff coordinates.

    Parameters:
    spacegroup: int
       int between 1 and 230
    wyckoffs: list
       wyckoff positions, for example ['a', 'a', 'b', 'c']
    species: list
       atomic species, for example ['Fe', 'O', 'O', 'O']
    """

    def __init__(self,
                 spacegroup=None,
                 wyckoffs=None,
                 species=None,
                 verbose=True
                 ):

        super().__init__(spacegroup=spacegroup,
                         wyckoffs=wyckoffs)

        self.spacegroup = spacegroup
        self.wyckoffs = wyckoffs
        self.species = species

        self.set_lattice_dof()
        if self.wyckoffs is not None:
            #self.p_name = self.get_prototype_name(species)
            self.parameter_guess = self.initial_guess()

        self.verbose = verbose

    def set_lattice_dof(self):
        """Set degrees of freedom for lattice constants and angles
        based on spacegroup
        """
        sg = self.spacegroup
        if sg in [1, 2]:  # Triclinic
            relax = [0, 1, 2]
            fixed_angles = {}

        elif 3 <= sg <= 15:  # Monoclinic
            relax = [0, 1, 2]
            fixed_angles = {'alpha': 90, 'gamma': 90}

        elif 16 <= sg <= 74:  # Orthorombic
            relax = [0, 1, 2]
            fixed_angles = {'alpha': 90, 'beta': 90,
                            'gamma': 90}

        elif 75 <= sg <= 142:  # Tetragonal
            relax = [[0, 1], 2]  # a = b
            fixed_angles = {'alpha': 90, 'beta': 90,
                            'gamma': 90}

        elif 143 <= sg <= 167:  # Trigonal
            relax = [[0, 1], 2]  # a = b
            fixed_angles = {'alpha': 90, 'beta': 90,
                            'gamma': 120}

        elif 168 <= sg <= 194:  # Hexagonal
            relax = [[0, 1], 2]  # a = b
            fixed_angles = {'alpha': 90, 'beta': 90,
                            'gamma': 120}

        elif 195 <= sg <= 230:  # Cubic
            relax = [[0, 1, 2]]  # a = b = c
            fixed_angles = {'alpha': 90, 'beta': 90,
                            'gamma': 90}

        self.lattice_variables = \
            np.array(['a', 'b', 'c'])[[r[0] if isinstance(r, list)
                                       else r for r in relax]]

        self.angle_variables = \
            [a for a in ['alpha', 'beta', 'gamma']
             if not a in fixed_angles]

        self.d_o_f = relax
        self.fixed_angles = fixed_angles

    def initial_guess(self):
        """ Rough estimate of parameters """

        parameter_guess = {}

        covalent_radii = [cradii[a_n[s]] for s in self.species]
        mean_radii = np.mean(covalent_radii)

        for i, lat_par in enumerate(self.lattice_variables):
            parameter_guess.update({lat_par: mean_radii * 4 * (1 + 0.1 * i)})

        if 'alpha' in self.angle_variables:
            parameter_guess.update({'alpha': 88})

        if 'beta' in self.angle_variables:
            parameter_guess.update({'beta': 94})

        if 'gamma' in self.angle_variables:
            parameter_guess.update({'gamma': 86})

        for i, p in enumerate(self.coor_variables):
            parameter_guess.update({p: rand(1)[0] * 0.99})

        self.parameter_guess = parameter_guess

        atoms = self.construct_atoms()

        self.natoms = atoms.get_number_of_atoms()

        parameter_guess.update({'a': mean_radii * 4 * self.natoms ** (1 / 3)})

        return parameter_guess

    def get_parameter_estimate(self,
                               master_parameters=None,
                               proximity=0.90,
                               max_candidates=1):
        """
        Optimize lattice parameters for prototype. 

        Parameters:
        master_parameters: dict
           fixed cell parameters and values that will not be optimized.
        proximity: float close to 1
           Proximity of atoms, in the relative sum of atomic radii.
        """

        cell_parameters = self.parameter_guess
        master_parameters = master_parameters or {}

        if master_parameters:
            cell_parameters.update(master_parameters)

        optimize_wyckoffs = (not np.all([c in master_parameters for c
                                         in self.coor_variables])
                             and len(self.coor_variables) > 0)

        optimize_angles = (not np.all([c in master_parameters
                                       for c in self.angle_variables])
                           and len(self.angle_variables) > 0)

        optimize_lattice = not np.all([c in master_parameters for
                                       c in self.lattice_variables])

        if not np.any([optimize_wyckoffs, optimize_angles, optimize_lattice]):
            print('No parameters to optimize!')
            return [cell_parameters]

        if optimize_wyckoffs or optimize_angles:
            opt_cell_parameters = []

            if self.verbose:
                print('Running ML + GA for {} - {} - {}. {} atoms'
                      .format(self.spacegroup, self.wyckoffs, self.species,
                              self.natoms))

            atoms_list = \
                self.run_ml_ga_optimization(
                    master_parameters=master_parameters,
                    optimize_lattice=optimize_lattice,
                    optimize_angles=optimize_angles,
                    optimize_wyckoffs=optimize_wyckoffs,
                    max_candidates=max_candidates)

            for atoms in atoms_list:
                opt_parameters = \
                    self.get_cell_parameters(
                        atoms)  # .update(master_parameters)
                opt_cell_parameters += [opt_parameters]

        elif optimize_lattice:
            # Scale fixed lattice points
            opt_atoms_list = []
            opt_cell_parameters = []
            if self.verbose:
                print('Optimizing lattice constants for {} - {} - {}. {} atoms'
                      .format(self.spacegroup, self.wyckoffs, self.species,
                              self.natoms))

            atoms = self.construct_atoms(cell_parameters)
            opt_atoms = \
                self.optimize_lattice_constants(atoms,
                                                optimize_wyckoffs=optimize_wyckoffs,
                                                view_images=False,
                                                proximity=proximity)

            opt_cell_parameters = [self.get_cell_parameters(opt_atoms)]
        else:
            opt_cell_parameters = cell_parameters_list
            opt_atoms_list = atoms_list

        return opt_cell_parameters

    def crossover(self, population, n_per_couple=10, n_children=None):
        """Generate new candidate by mixing values of two"""
        children = []
        variables = list(population[0].keys())
        for i, pop in enumerate(population):
            for j, pop2 in enumerate(population):
                if i == j or np.all(np.isclose(sorted(pop.values()),
                                               sorted(pop2.values()),
                                               rtol=0.05)):  # inbred child
                    continue
                for m in range(n_per_couple):
                    child = {}
                    for p in variables:
                        c_point = rand(1)[0]  # 0.5 + (rand(1)[0] - 0.5) * 0.5
                        value = pop[p] * c_point + pop2[p] * (1 - c_point)
                        child.update({p: value})

                    children += [child]
        n = n_children or len(population)
        shuffle(children)
        return children[:n]

    def mutation(self, population, n_per_pop=10, n_children=None):
        mutated = []
        variables = list(population[0].keys())
        for i, pop in enumerate(population):
            for m in range(n_per_pop):
                mutant = pop.copy()
                indices = randint(len(variables), size=len(variables) // 2)

                for j in indices:
                    p = variables[j]
                    value = pop[p]
                    if p in self.coor_variables:
                        value += (rand(1)[0] - 0.5) * 0.1
                    elif p in self.angle_variables:
                        value = (rand(1)[0] - 0.5) * 30 + 90
                    mutant.update({p: value})
                mutated += [mutant]
            n = n_children or len(population)
        shuffle(mutated)
        return mutated[:n]

    def run_ml_ga_optimization(self,
                               master_parameters=None,
                               optimize_lattice=True,
                               optimize_angles=True,
                               optimize_wyckoffs=True,
                               use_fitness_sharing=False,
                               batch_size=6,
                               max_candidates=1,
                               debug=False):
        """
        ML-accelerated Genetic algorithm optimization 
        for free wyckoff coordinates and lattice angles.
        """

        cell_parameters = self.initial_guess()
        if master_parameters:
            cell_parameters.update(master_parameters)
        population = []

        feature_variables = []
        if optimize_wyckoffs:
            feature_variables += self.coor_variables

        if optimize_angles:
            feature_variables += self.angle_variables

        n_parameters = len(feature_variables)
        population_size = min([5000, n_parameters * 100])
        population = []

        test_features = []
        for n in range(population_size):
            t_f = []
            parameters = {}
            if optimize_wyckoffs:
                for i, p in enumerate(self.coor_variables):
                    val = rand(1)[0] * 0.99
                    parameters.update({p: val})
                    t_f += [val]

            if optimize_angles:
                for i, p in enumerate(self.angle_variables):
                    val = (rand(1)[0] - 0.5) * 30 + 90
                    parameters.update({p: val})
                    t_f += [val]
            population += [parameters]
            test_features += [t_f]

        atoms = self.construct_atoms()
        covalent_radii = np.array([cradii[n] for n in atoms.numbers])
        M = covalent_radii * np.ones([len(atoms), len(atoms)])
        self.min_distances = (M + M.T)

        primitive_voronoi = len(atoms) > 64

        covalent_radii = np.array([cradii[n] for n in atoms.numbers])
        covalent_volume = np.sum(4/3 * np.pi * covalent_radii ** 3)
        cell_length = (covalent_volume * 2) ** (1/3)

        test_features = np.array(test_features)
        batch_indices = np.random.randint(len(test_features),
                                          size=batch_size)

        train_features = None

        fitness = np.array([])
        all_structures = []

        converged = False
        iter_id = 1
        train_population = []

        while not converged:
            bad_indices = []
            for i in batch_indices:
                pop = population[i]
                train_population += [pop]
                cell_parameters.update(pop)
                atoms = self.construct_atoms(cell_parameters)
                atoms = self.optimize_lattice_constants(atoms,
                                                        proximity=0.9,
                                                        optimize_wyckoffs=False)
                if atoms is None:
                    bad_indices += [i]
                    continue
                parameters = cell_to_cellpar(atoms.get_cell())

                if primitive_voronoi:
                    # Use primitive cell for voronoi analysis for
                    # large systems
                    for i, param_name in enumerate(['a', 'b', 'c',
                                                    'alpha', 'beta', 'gamma']):
                        if param_name in cell_parameters:
                            cell_parameters.update({param_name: parameters[i]})

                    atoms = self.construct_atoms(cell_parameters,
                                                 primitive_cell=True)

                fit = get_fitness(atoms)
                connections = None
                if fit > -2:
                    connections = get_connections(atoms, decimals=1)

                fitness = np.append(fitness, fit)
                all_structures += [{'parameters': cell_parameters,
                                    'atoms': atoms.copy(),
                                    'fitness': fit,
                                    'graph': connections}]

            best_fitness = np.max(fitness)
            print('iter {} best_fitnes:'.format(iter_id),  np.max(fitness))
            batch_indices = [idx for idx in batch_indices if not idx
                             in bad_indices]
            if train_features is None:
                train_features = test_features[batch_indices]
            else:
                train_features = np.append(train_features,
                                           test_features[batch_indices],
                                           axis=0)

            test_features = np.delete(test_features, batch_indices, axis=0)
            test_features = np.delete(test_features, bad_indices, axis=0)

            population = np.delete(population, batch_indices, axis=0)

            indices = np.argsort(fitness)[::-1]
            ga_survived = np.array(train_population)[indices][:10]
            new_population = self.crossover(ga_survived, n_children=50) + \
                self.mutation(ga_survived, n_children=10)

            for pop in new_population:
                val = []
                for v in feature_variables:
                    val += [pop[v]]
                val = np.expand_dims(val, axis=0)

                test_features = np.append(test_features, val, axis=0)
                population = np.append(population, pop)

            features, bad_indices = \
                clean_features({'train': train_features,
                                'test': test_features})

            fitness = np.delete(fitness, bad_indices['train'])
            test_features = np.delete(test_features, bad_indices['test'],
                                      axis=0)
            train_features = np.delete(train_features, bad_indices['train'],
                                       axis=0)
            population = np.delete(population, bad_indices['train'])

            try:
                Model = get_regression_model('catlearn')(
                    features['train'],
                    np.array(fitness),
                    optimize_hyperparameters=True,
                    kernel_width=1,
                    #bounds=((0.5, 5),)
                )
            except:
                print('Fixed params')
                Model = get_regression_model('catlearn')(
                    features['train'],
                    np.array(fitness),
                    optimize_hyperparameters=False,
                    kernel_width=3)

            result = Model.predict(features['test'])
            predictions = result['prediction']
            unc = result['uncertainty']

            # Acquisition function mix two acquisition functions
            AQU1 = predictions + 0.1 * unc
            AQU2 = predictions + 0.5 * unc

            indices1 = np.argsort(AQU1)[::-1][:batch_size]
            indices2 = np.argsort(AQU2)[::-1][:batch_size]

            indices2 = np.array([int(aqu_i)
                                 for aqu_i in indices2 if not aqu_i in indices1])

            bs1 = batch_size // 2
            bs2 = batch_size - bs1
            batch_indices = np.array(
                np.append(indices1[:bs1], indices2[:bs2]), dtype=int)

            iter_id += 1

            if debug:
                import pylab as p
                idx = np.argsort(predictions)
                p.plot(range(len(predictions)), predictions[idx])
                p.plot(range(len(predictions)),
                       predictions[idx] + unc[idx], '--')
                p.plot(range(len(predictions)),
                       predictions[idx] * 0 + best_fitness, '--')
                p.show()

            if iter_id > 30 or len(predictions) < 7:
                converged = True
            elif not np.max(AQU1) > best_fitness and iter_id > 5:
                converged = True
            # elif best_fitness > 0.95:
            #    converged = True

        indices = np.argsort(fitness)[::-1][:max_candidates]
        fitness = fitness[indices]

        all_structures = np.array(all_structures)[indices]
        all_graphs = np.array([s['graph'] for s in all_structures])
        f_max = max(fitness)

        if fitness[0] < 0.8:
            indices = [0]
        else:
            indices = [i for i, f in enumerate(fitness) if
                       f > 0.8 * f_max and not
                       np.any(all_graphs[i] in all_graphs[:i])
                       ]

        if primitive_voronoi:
            # generate conventional structure
            all_atoms = [self.construct_atoms(all_structures[i]['parameters'])
                         for i in indices]
        else:
            all_atoms = [s['atoms'] for s in all_structures]
            all_atoms = [all_atoms[i] for i in indices]
        print('final fitness: ', fitness[indices])
        return all_atoms

    def construct_atoms(self, cell_parameters=None, primitive_cell=False):
        if cell_parameters:
            self.parameter_guess.update(cell_parameters)

        relative_positions = []
        symbols = []

        for i_w, w in enumerate(self.wyckoffs):
            m = self.wyckoff_multiplicities[w]
            w_arrays = self.wyckoff_symmetries[w]

            wyckoff_coordinate = []
            for direction in ['x', 'y', 'z']:
                wyckoff_coordinate += [
                    self.parameter_guess.get(direction + w + str(i_w), 0)]

            for w_array in w_arrays:
                M = w_array[:, :3]
                c = w_array[:, 3]
                relative_positions += [
                    tuple(np.dot(wyckoff_coordinate, M.T) + c)]

            symbols += [self.species[i_w]] * m
        cellpar = []
        for p in ['a', 'b', 'c']:
            value = self.parameter_guess.get(p, None)
            if value is None:
                value = self.parameter_guess.get('a', 4)
            cellpar += [value]

        for angle in ['alpha', 'beta', 'gamma']:
            value = self.parameter_guess.get(angle, None)
            if not value:
                value = self.fixed_angles.get(angle)
            cellpar += [value]

        atoms = crystal(tuple(symbols),
                        relative_positions,
                        spacegroup=self.spacegroup,
                        cellpar=cellpar,
                        primitive_cell=primitive_cell,
                        onduplicates='keep',
                        symprec=1e-5)

        # if not primitive_cell:
        #    #print(len(relative_positions), len(atoms))
        #    if not len(relative_positions) == len(atoms):
        #        return None
        return atoms

    """
    def check_prototype(self, atoms):
        #Check that spacegroup and wyckoff positions did not change

        PC = PrototypeClassification(atoms)

        prototype = PC.get_classification(include_parameters=False)

        if not self.spacegroup == prototype['spacegroup']:
            print('Spacegroup changed')
            return False

        elif not (self.wyckoffs == prototype['wyckoffs']
                  and self.species == prototype['species']):
            print('Wyckoff positions changed')
            return False
        else:
            return True
    """

    def optimize_lattice_constants(self,
                                   atoms,
                                   optimize_wyckoffs=False,
                                   view_images=False,
                                   proximity=0.95):
        """
        Optimize lattice parameters by reducing the cell size (one direction at
        the time) until atomic distances on the closest pair reaches the
        sum of the covalent radii.

        if the optimize_wyckoffs parameter is set, the wyckoff coordinates 
        are optimized together with the lattice constants. 
        """
        if not atoms:
            return atoms

        cell = atoms.cell

        Dm, distances = get_interatomic_distances(atoms)
        covalent_radii = np.array([cradii[n] for n in atoms.numbers])
        M = covalent_radii * np.ones([len(atoms), len(atoms)])

        self.min_distances = (M + M.T) * proximity

        # scale up or down
        images = [atoms.copy()]
        soft_limit = 1.36
        scale = np.min(distances / self.min_distances / soft_limit)
        atoms.set_cell(atoms.cell * 1 / scale, scale_atoms=True)

        Dm, distances = get_interatomic_distances(atoms)
        images += [atoms.copy()]

        self.hard_limit = soft_limit
        increment = 0.99

        while self.hard_limit > 1:
            cell_norm = np.linalg.norm(cell, axis=1)
            axis_length_ix = np.argsort(
                np.mean(cell_norm[d]) for d in self.d_o_f)
            if optimize_wyckoffs:
                images += self.run_wyckoff_optimization_loop(atoms)
                atoms = images[-1].copy()
                Dm, distances = get_interatomic_distances(atoms)
                cell_volume = atoms.get_volume()

            min_distances = np.unravel_index(
                np.argmin(distances/self.min_distances),
                distances.shape)
            min_dir_vector = Dm.copy()[min_distances]
            min_dir_vector /= distances[min_distances]
            dot_directions = []
            for direction in self.d_o_f:
                v = np.zeros([3])
                v[direction] = 1
                v[direction] /= np.linalg.norm(v[direction])
                dot_directions += [np.abs(np.dot(v, min_dir_vector))]

            direction_order = np.argsort(dot_directions)[::-1]

            for direction in np.array(self.d_o_f)[direction_order]:
                self.hard_limit *= increment
                while np.all(distances > self.min_distances * self.hard_limit):
                    cell = atoms.cell.copy()
                    cell[direction, :] *= increment
                    atoms.set_cell(cell, scale_atoms=True)
                    Dm, distances = get_interatomic_distances(atoms.copy(),
                                                              Dm.copy(),
                                                              increment,
                                                              direction)

                    cell_volume = atoms.get_volume()
                images += [atoms.copy()]

        if view_images:
            view(images)

        return images[-1]

    def get_move_pairs(self, distances, cutoff=1.1):
        move_indices = np.nonzero(
            distances / self.min_distances < self.hard_limit * cutoff)
        move_pairs = []
        if not len(move_indices[0]) > 0:
            return []

        for m in range(len(move_indices[0])):
            pair = sorted([move_indices[0][m], move_indices[1][m]])
            if np.any([self.free_atoms[a] for a in pair]) and pair[0] != pair[1]:
                move_pairs += [pair]
        if not move_pairs:
            return []

        move_pairs = np.unique(move_pairs, axis=0)
        clean_move_pairs = []
        del_pairs = []
        for i, pair in enumerate(move_pairs):
            sym_pairs = []
            for a in self.symmetry_map[pair[0]]:
                for b in self.symmetry_map[pair[1]]:
                    sym_pairs += [sorted([a, b])]
            if not np.any([sym_pair in move_pairs[:i] for sym_pair in sym_pairs]):
                clean_move_pairs += [pair]

        return clean_move_pairs

    def run_wyckoff_optimization_loop(self, atoms):
        Dm, distances = get_interatomic_distances(atoms)
        relative_distances = distances / self.min_distances
        move_pairs = self.get_move_pairs(distances, 1.1)
        atoms_distance = 0.90
        niter = 0
        images = [atoms.copy()]
        while len(move_pairs) > 0 and niter < atoms.get_number_of_atoms() * 3:
            for a1, a2 in move_pairs:
                transform_vector = Dm[a1][a2]
                transform_vector /= np.linalg.norm(transform_vector)

                move = -np.abs(1.02 - relative_distances[a1][a2]) * \
                    self.min_distances[a1][a2] * \
                    atoms_distance * self.hard_limit
                transform_vector *= move

                # Project onto wyckoff coordinate vector
                transform_vector1 = self.get_wyckoff_transform_vector(
                    transform_vector, a1, a1, atoms.cell)

                atoms[a1].position += transform_vector1

                if not a2 in self.symmetry_map[a1]:
                    # No symmetry restrictions bewteen atom1 and atom2
                    transform_vector2 = self.get_wyckoff_transform_vector(
                        -transform_vector, a2, a2, atoms.cell)
                    atoms[a2].position += transform_vector2
                    sym_iter = [a1, a2]
                    trans_vectors = [transform_vector1, transform_vector2]
                else:
                    sym_iter = [a1]
                    trans_vectors = [transform_vector1]
                for i, ia in enumerate(sym_iter):
                    # Move all atoms in same wyckoff position as atom1 and atom2
                    co_trans = self.symmetry_map[ia].copy()
                    if ia in co_trans:
                        co_trans.remove(ia)

                    for ic in co_trans:
                        transform_vector_w = \
                            self.get_wyckoff_transform_vector(
                                trans_vectors[i], ia, ic, atoms.cell)
                        atoms[ic].position += transform_vector_w

            atoms.wrap()
            images += [atoms.copy()]
            cell_volume = atoms.get_volume()
            Dm, distances = get_interatomic_distances(atoms)
            relative_distances = distances / self.min_distances
            atoms_distance *= 0.90
            move_pairs = self.get_move_pairs(distances, 1.1 * atoms_distance)
            niter += 1
        return images


def clean_parameter_input(cell_parameters):
    if 'a' in cell_parameters:
        a = cell_parameters['a']
        for l_c in ['b', 'c']:
            if l_c in cell_parameters:
                l = cell_parameters[l_c]
                del cell_parameters[l_c]
                cell_parameters.update({l_c + '/a': l/a})

    return cell_parameters


def get_interatomic_distances(atoms, D=None, scale=None, direction=None):
    if D is not None:
        D[:, :, direction] *= scale
        D.shape = (-1, 3)
        distances = np.sqrt((D**2).sum(1))
        D.shape = (-1, len(atoms), 3)
        distances.shape = (-1, len(atoms))

    else:
        D, distances = get_distances(atoms.positions,
                                     cell=atoms.cell, pbc=True)

    min_cell_width = np.min(np.linalg.norm(atoms.cell, axis=1))
    min_cell_width *= np.ones(len(atoms))
    np.fill_diagonal(distances, min_cell_width)

    return D, distances



import os
import sys
import subprocess
import json
import numpy as np
import pandas as pd
import ase.db
from ase.db.sqlite import SQLite3Database
from ase.io import read
import sqlite3

from protosearch.utils import get_tri_basepath
from protosearch.utils.standards import VaspStandards

init_commands = [
    """CREATE TABLE prototype (
    id INTEGER PRIMARY KEY AUTOINCREMENT,
    name text UNIQUE,
    natom int,
    spacegroup int,
    npermutations int,
    permutations text,
    parameters text,
    species text,
    wyckoffs text,
    source text);""",

    """CREATE TABLE prediction (
    batch_no,
    id integer,
    energy real,
    uncertainty real,
    FOREIGN KEY (id) REFERENCES systems(id));
    """,

    """CREATE TABLE enumeration (
    stoichiometry text,
    spacegroup int,
    number int,
    num_type text
    );""",

    """CREATE TABLE batch_status (
    batch_no INTEGER PRIMARY KEY AUTOINCREMENT,
    structure_ids text,
    completed_ids text
    );""",

    """CREATE TABLE status (
    id integer,
    chemical_formulas text,
    batch_size int,
    enumerated int,
    fingerprinted int,
    initialized int,
    last_batch_no int,
    n_completed int,
    n_errored int,
    most_stable_id int
    );""",

    """INSERT into status
    (id, enumerated, fingerprinted, initialized, n_completed, n_errored)
    VALUES
    (0, 0, 0, 0, 0, 0);"""
]

columns = ['id', 'name', 'natom', 'spacegroup', 'npermutations', 'permutations',
           'parameters', 'species', 'wyckoffs']

json_columns = [5, 6, 7, 8]


class PrototypeSQL:
    def __init__(self,
                 filename=None,
                 stdin=sys.stdin,
                 stdout=sys.stdout):

        if filename:
            assert filename.endswith(
                '.db'), 'filename should have .db extension'
        else:
            basepath = get_tri_basepath()
            filename = basepath + '/prototypes.db'
        self.filename = filename
        self.initialized = False
        self.default = 'NULL'
        self.connection = None
        self.stdin = stdin
        self.stdout = stdout
        self.ase_db = ase.db.connect(self.filename)

    def _connect(self):
        #self.ase_db = ase.db.connect(self.filename)
        return sqlite3.connect(self.filename, timeout=600)

    def __enter__(self):
        """Set connection upon entry using with statement"""
        assert self.connection is None
        self.connection = self._connect()
        return self

    def __exit__(self, exc_type, exc_value, tb):
        """Commit changes upon exit"""
        if exc_type is None:
            self.connection.commit()
        else:
            self.connection.rollback()
        self.connection.close()
        self.connection = None

    def _initialize(self, con):
        """Set up tables in SQL"""
        if self.initialized:
            return

        SQLite3Database()._initialize(con)  # ASE db initialization

        cur = con.execute(
            'SELECT COUNT(*) FROM sqlite_master WHERE name="status"')

        if cur.fetchone()[0] == 0:  # no prototype table
            for init_command in init_commands:
                print(init_command)
                try:
                    con.execute(init_command)  # Create tables
                except:
                    pass
            con.commit()

        self.initialized = True

    def get_status(self):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        cur.execute('SELECT * from status;')

        status = cur.fetchall()[0]
        status_dict = {
            'chemical_formulas': status[1],
            'batch_size': status[2],
            'enumerated': status[3],
            'fingerprinted': status[4],
            'initialized': status[5],
            'last_batch_no': status[6],
            'n_completed': status[7],
            'n_errored': status[8],
            'most_stable_id': status[9]}
        return status_dict

    def write_status(self, **args):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        for key, value in args.items():
            query = "UPDATE status SET {}=".format(key)

            if isinstance(value, str):
                query += "'{}'".format(value)
            elif isinstance(value, list):
                query += "'{}'".format(json.dumps(value))
            else:
                query += "{}".format(value)
            query += ' where id=0'
            cur.execute(query)

        if not self.connection:
            con.commit()
            con.close()
        return

    def write_job_status(self):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        cur.execute(
            "SELECT count(id) from number_key_values where key='relaxed' and value=1")
        n_completed = cur.fetchall()[0][0]
        cur.execute(
            "SELECT count(id) from number_key_values where key='completed' and value=-1")
        n_errored = cur.fetchall()[0][0]

        self.write_status(n_completed=n_completed, n_errored=n_errored)

        if not self.connection:
            con.commit()
            con.close()
        return

    def write_prototype(self, entry):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        species = []

        permutations = entry['specie_permutations']

        values = (entry['name'],
                  entry['natom'],
                  entry['spaceGroupNumber'],
                  len(permutations),
                  json.dumps(permutations),
                  json.dumps(entry['parameters']),
                  json.dumps(entry['species']),
                  json.dumps(entry['wyckoffs']),
                  entry['source']
                  )

        q = self.default + ',' + ', '.join('?' * len(values))
        cur.execute('INSERT or IGNORE INTO prototype VALUES ({})'.format(q),
                    values)
        if not self.connection:
            con.commit()
            con.close()

        return

    def write_batch_status(self, batch_no, structure_ids):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        structure_ids = json.dumps(structure_ids)

        cur.execute("INSERT into batch_status (batch_no, structure_ids) VALUES ({}, '{}')".
                    format(batch_no, structure_ids))

        if not self.connection:
            con.commit()
            con.close()

        return

    def is_enumerated(self, stoichiometry, spacegroup, number, num_type):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        cur.execute(
            """SELECT count(*) from enumeration where stoichiometry='{}' and spacegroup={} and
            number={} and num_type='{}';""".format(stoichiometry, spacegroup, number, num_type))

        count = cur.fetchall()[0]

        return count

    def write_enumerated(self, stoichiometry, spacegroup, number, num_type):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        cur.execute(
            """INSERT into enumeration (stoichiometry, spacegroup, number, num_type)
            VALUES ('{}', {}, {}, '{}')""".
            format(stoichiometry, spacegroup, number, num_type)
        )

        if not self.connection:
            con.commit()
            con.close()
        return

    def update_status_complete(self, batch_no, structure_ids):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        structure_ids = json.dumps(structure_ids)

        cur.execute("UPDATE status SET batch_no={}, structure_ids='{}' where id = 0".
                    format(batch_no, structure_ids))

        if not self.connection:
            con.commit()
            con.close()

        return

    def select_atoms(self, **kwargs):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        atoms_list = []
        for row in self.ase_db.select(**kwargs):
            atoms_list += row.toatoms()

        return atoms_list

    def get_atom_by_id(self, id):
        con = self.connection or self._connect()
        self._initialize(con)
        row = self.ase_db.get(id=id)
        return row.toatoms()

    def get_atoms_list(self, ids):
        con = self.connection or self._connect()
        self._initialize(con)
        atoms_list = []
        for id in ids:
            atoms_list += [self.ase_db.get(id=int(id)).toatoms()]
        return atoms_list

    def select(self, **kwargs):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        query = ''
        for key, value in kwargs.items():
            if value is None:
                continue
            if len(query) > 0:
                query += ' and'
            if key == 'max_atoms':
                query += "natom<={}".format(value)
            elif key == 'spacegroups':
                value = [str(v) for v in value]
                sg_str = '(' + ','.join(value) + ')'
                query += " spacegroup in {}".format(sg_str)
            elif isinstance(value, str):
                query += " {}='{}'".format(key, value)
            else:
                query += ' {}={}'.format(key, value)
        statement = 'SELECT * from prototype'
        if query:
            statement += ' where {}'.format(query)

        cur.execute(statement)
        data = cur.fetchall()

        result = []
        for l in range(len(data)):
            result += [{}]
            for i, k in enumerate(columns):
                if i in json_columns:
                    result[l].update({k: json.loads(data[l][i])})
                else:
                    result[l].update({k: data[l][i]})

        return result

    def get_prototype_id(self, p_name):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        cur.execute('SELECT id from prototype where name={};'.format(p_name))
        name = cur.fetchall()[0]

        return name

    def get_prototype_info(self, p_name):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        cur.execute(
            'SELECT spacegroup,wyckoffs,from prototype where name={};'.format(p_name))
        name = cur.fetchall()[0]

        return name

    def is_calculated(self, formula, p_name):
        con = self.connection or self._connect()
        if self.ase_db.count('completed>-1',
                             submitted=1,
                             formula=formula,
                             p_name=p_name) > 0:
            print('Allready calculated')
            return True
        else:
            return False

    def is_calculated_id(self, id):
        con = self.connection or self._connect()
        if self.ase_db.count('completed>-1',
                             id=id,
                             p_name=p_name) > 0:
            print('Allready calculated')
            return True
        else:
            return False

    def get_next_batch_no(self):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        cur.execute(
            "SELECT max(value) from number_key_values where key='batch'")
        cur_batch = cur.fetchall()
        if cur_batch:
            if cur_batch[0][0]:
                return cur_batch[0][0] + 1
            else:
                return 0
        else:
            return 0

    def get_structure_ids(self, start_id=1, n_ids=None):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        query = "SELECT id from systems where id>{}".format(start_id-1)
        if n_ids:
            query += ' limit {}'.format(n_ids)
        cur.execute(query)
        ids = cur.fetchall()
        ids = [i[0] for i in ids]
        return ids

    def get_completed_structure_ids(self, max_energy=0):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        query = \
            """select distinct id from number_key_values
            where key='relaxed' and value=1
            and id not in (SELECT distinct id from systems where energy > {})
            order by id""".format(max_energy)
        cur.execute(query)
        ids = cur.fetchall()
        ids = [i[0] for i in ids]
        return ids

    def get_uncompleted_structure_ids(self, unsubmitted_only=True):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        if unsubmitted_only:
            query =\
                """SELECT distinct id from number_key_values
                where key='submitted' and value=0 and id not in
                (SELECT distinct value from number_key_values
                where key='initial_id') order by id"""
        else:
            query =\
                """SELECT distinct id from number_key_values
                where key='completed' and value=0 and id not in
                (SELECT distinct value from number_key_values
                where key='initial_id') order by id"""

        cur.execute(query)
        ids = cur.fetchall()
        ids = [i[0] for i in ids]
        return ids

    def get_initial_structure_ids(self, completed=False):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        if completed:
            query =\
                """select distinct id from number_key_values
                where key='relaxed' and value=0 and id in 
                (select distinct id from number_key_values
                where key='completed' and value=1)
                order by id"""
        else:
            query =\
                """select distinct id from number_key_values
                where key='relaxed' and value=0
                order by id"""
        cur.execute(query)
        ids = cur.fetchall()
        ids = [i[0] for i in ids]
        return ids

    def write_dataframe(self, table, df):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        ids = df['id'].values
        id_string = ', '.join([str(i) for i in ids])
        cur.execute(
            """SELECT count(*) FROM sqlite_master WHERE
            type='table' AND name='{}';""".format(table))
        table_count = cur.fetchall()[0][0]
        if table_count:
            cur.execute(
                'DELETE from {} where id in ({})'.format(table, id_string))
        df.to_sql(table, con=con, index=False,
                  index_label=None, if_exists='append')

    def load_dataframe(self, table, ids=None):
        con = self.connection or self._connect()
        self._initialize(con)

        query = 'SELECT * from {}'.format(table)
        if ids is not None:
            ids = sorted(ids)
            id_str = ','.join([str(i) for i in ids])
            query += ' where id in ({})'.format(id_str)
        query += ' order by id'
        df = pd.read_sql_query(query, con)
        return df

    def write_fingerprint(self, id, input_data, output_data=None):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        cur.execute('DELETE from fingerprint where id={}'.format(id))

        input_data = json.dumps(input_data.tolist())
        columns = ['id', 'input']
        values = [id, input_data]

        if output_data:
            output_data = json.dumps(output_data)
            columns += ['output']
            values += [output_data]

        values = ['{}'.format(v) if isinstance(v, int) else "'{}'".format(v)
                  for v in values]

        columns = ','.join(columns)
        values = ','.join(values)
        cur.execute(
            'INSERT INTO fingerprint ({}) VALUES ({})'.format(columns, values))
        con.commit()
        con.close()

    def get_new_fingerprint_ids(self, completed=True):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        cur.execute(
            """SELECT count(*) FROM sqlite_master WHERE
            type='table' AND name='target';""")
        table_count = cur.fetchall()[0][0]

        if completed:
            query = 'SELECT id from systems where energy is not null'
            if table_count:
                query += ' and id not in (SELECT distinct id from target)'
        else:
            query = \
                """SELECT id from systems"""
            if table_count:
                query += ' where id not in (SELECT distinct id from fingerprint)'
        cur.execute(query)
        ids = cur.fetchall()
        ids = [i[0] for i in ids]
        return ids

    def write_predictions(self, batch_no, ids, Efs, var):
        # Just save formation energy and uncertainty for now
        # link to structure id
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()

        cur.execute('DELETE FROM prediction WHERE batch_no={}'.format(batch_no))
        for i, idi in enumerate(ids):
            cur.execute('INSERT INTO prediction VALUES ({}, {}, {}, {})'.format(
                batch_no, idi, float(Efs[i]), float(var[i])))

        con.commit()
        con.close()

    def get_predictions(self, batch_no=None):
        con = self.connection or self._connect()
        self._initialize(con)
        cur = con.cursor()
        if not batch_no:
            cur.execute('SELECT distinct batch_no FROM prediction')
            batch_nos = cur.fetchall()
            batch_nos = [b[0] for b in batch_nos]
        else:
            batch_nos = [batch_no]

        predictions = []
        for b in batch_nos:
            cur.execute(
                'SELECT * FROM prediction where batch_no={};'.format(b))
            data = cur.fetchall()
            ids = [d[1] for d in data]
            energies = np.array([float(d[2]) for d in data])
            var = np.array([float(d[3]) for d in data])
            predictions += [{'batch_no': b,
                             'ids': ids,
                             'energies': energies,
                             'vars': var}]
        return predictions

    def get_pandas_tables(self,
                          write_csv_tables=False,
                          tables_list=None):
        """Convert SQL tables to Pandas dataframes.

        Parameters:
        ----------
        write_csv_tables: Bool
            Write data tables to .csv files to current working dir
        tables_list: list or None
            List of table names to retrieve if you don't want all of them
            tables_list = [
                'systems',
                'sqlite_sequence',
                'species',
                'keys',
                'text_key_values',
                'number_key_values',
                'information',
                'prototype',
                'fingerprint',
                'target'
                'prediction',
                'enumeration',
                'batch_status',
                'status']
        """
        db = sqlite3.connect(self.filename)
        cursor = db.cursor()
        cursor.execute(
            "SELECT name FROM sqlite_master WHERE type='table';")
        tables = cursor.fetchall()

        tables_dict = {}
        for table_name in tables:
            table_name = table_name[0]

            if tables_list is not None:
                if table_name in tables_list:
                    table_i = get_table(table_name, db, write_csv_tables)
                    tables_dict[table_name] = table_i
            else:
                table_i = get_table(table_name, db, write_csv_tables)
                tables_dict[table_name] = table_i

        cursor.close()
        db.close()

        # tables = to_csv(db_file=self.filename)
        return(tables_dict)


def get_table(table_name, db, write_csv_tables):
    table = pd.read_sql_query("SELECT * from %s" % table_name, db)
    # tables_dict[table_name] = table
    if write_csv_tables:
        table.to_csv(table_name + '.csv', index_label='index')
    return(table)



import numpy as np
import pylab as p
import ase


from protosearch.active_learning.active_learning import ActiveLearningLoop
from protosearch.ml_modelling.fingerprint import clean_features


class MetaAnalysis(ActiveLearningLoop):

    def __init__(self,
                 *args, **kwargs,
                 ):
        self.energies = None

        super().__init__(*args, **kwargs)

    def plot_fingerprint_variation(self):
        test_ids = self.DB.get_initial_structure_ids()
        all_fingerprints = self.DB.load_dataframe('fingerprint', ids=test_ids)
        all_fingerprints = {'train': all_fingerprints.values, 'test': None}
        all_fingerprints, bad_ids = clean_features(all_fingerprints,
                                                   scale=True)
        all_fingerprints = all_fingerprints['train']
        max_values = np.max(all_fingerprints, axis=0)
        indices = np.argsort(max_values)
        p.plot(all_fingerprints[:, indices].T)

        p.xlabel('Feature id')
        p.ylabel('Standardized feature value')
        p.title('Fingerprints for {} structures'.format(len(all_fingerprints)))
        p.show()

    def plot_prediction(self, energies, uncertainties, calculated_indices=[]):
        p.figure()

        x = np.arange(len(energies))
        p.plot(x, energies, 'x', label='energies')
        p.plot(x, energies - uncertainties, 'k--', label='std deviation')
        p.plot(x, energies + uncertainties, 'k--')

        new_calculated_indices = \
            [i for i in np.where(uncertainties == 0)[0]
             if not i in calculated_indices]

        p.plot(x[calculated_indices], energies[calculated_indices],
               'yo', label='calculated')
        if len(new_calculated_indices) > 0:
            p.plot(x[new_calculated_indices],
                   energies[new_calculated_indices], 'ro',
                   label='new calculated')

        p.xlabel('Structure #')
        p.ylabel('Energy(eV)')
        p.legend()

        return p

    def collect_results(self):
        energies = np.append(self.energies, self.targets)
        uncertainties = np.append(self.uncertainties,
                                  np.zeros_like(self.targets))

        indices = np.argsort(energies)
        energies = energies[indices]
        uncertainties = uncertainties[indices]

        calculated_ids = np.where(uncertainties == 0)[0]

        return energies, uncertainties, calculated_ids

    def get_new_prediction(self):
        self.train_ids = self.DB.get_completed_structure_ids()
        self.test_ids = \
            self.DB.get_uncompleted_structure_ids(unsubmitted_only=False)
        self.get_ml_prediction()

        energies, uncertainties, calculated_ids = self.collect_results()

        p = self.plot_prediction(energies,
                                 uncertainties,
                                 calculated_ids)

        p.title('Energy predictions')

        p.show()

    def plot_acquisition(self, kappa=0.5, method='LCB'):
        #self.train_ids = self.train_ids[:100]
        # if self.energies is None:

        self.train_ids = self.DB.get_completed_structure_ids()[:30]
        # get_completed_structure_ids(completed=0)
        self.test_ids = self.DB.get_structure_ids()

        self.test_ids = [i for i in self.test_ids if i not in self.train_ids]
        self.get_ml_prediction()

        all_ids = self.test_ids + self.train_ids
        all_energies = np.array(list(self.energies) + list(self.targets))
        all_uncertainties = np.array(list(self.uncertainties) +
                                     list(np.zeros_like(self.targets)))

        idx = np.argsort(all_energies)
        energies = all_energies[idx]
        uncer = all_uncertainties[idx]
        ids = [all_ids[i] for i in idx]
        p.plot(range(len(idx)), energies, 'x', label='energies')
        p.plot(range(len(idx)), energies -
               uncer / 2, 'k--', label='confidence')
        p.plot(range(len(idx)), energies +
               uncer / 2, 'k--')
        marker = ['o', '*', '<', 'D', 's', '^']

        self.acquire_batch(method=method, kappa=kappa, batch_size=10)
        acquisition_values = np.array(
            list(self.acquisition_values) + [None] * len(self.targets))
        p.plot(range(len(idx)), acquisition_values[idx], 'r--')
        new_ids_idx = [ids.index(i) for i in self.batch_ids]
        p.plot(new_ids_idx, energies[new_ids_idx],
               marker[0], label='{}'.format(method + str(kappa)))
        p.legend()
        p.show()

    def plot_predictions(self):
        #predictions = self.DB.get_predictions()
        predictions = []
        calculated = []
        if self.energies is None:
            self.get_new_prediction()

        energies, uncertainties, calculated_ids = self.collect_results()
        prediction = {'energies': energies,
                      'vars': uncertainties,
                      'batch_no': predictions[-1]['batch_no'] + 1}
        predictions += [prediction]

        for prediction in predictions[-5:]:
            p.figure()
            idx = np.argsort(prediction['energies'])
            energies = prediction['energies'][idx]
            unc = prediction['vars'][idx]
            self.plot_prediction(energies, inc)

            new_ids_idx = [ids.index(i) for i in new_ids]

            p.plot(old_ids_idx, energies[old_ids_idx],
                   'yo', label='previous batches')
            p.plot(new_ids_idx,
                   energies[new_ids_idx], 'ro', label='latest batch')

            p.xlabel('Structure id')
            p.title('Predictions for batch {}'.format(prediction['batch_no']))

            p.ylabel('Energy(eV)')
            p.legend()

            p.savefig('prediction_batch_{}.png'.format(prediction['batch_no']))

        p.show()

    def plot_test_predictions(self, method='random'):
        predictions = []
        calculated = []
        i = 0
        ids_old = []
        for prediction in self.test_run(acquire=method, kappa=0.5):
            i += 1
            p.figure()
            idx = np.argsort(prediction['energies'])
            energies = prediction['energies'][idx]
            unc = prediction['vars'][idx]
            ids = [int(prediction['ids'][i]) for i in idx]

            p.plot(range(len(idx)), energies, 'x', label='predictions')
            p.plot(range(len(idx)), energies -
                   unc / 2, '-', color='lightslategray', label='uncertainty')
            p.plot(range(len(idx)), energies + unc /
                   2, '-', color='lightslategray')

            idx1 = np.where(unc == 0)[0]
            ids1 = [ids[i] for i in idx1]

            new_ids = [i for i in ids1 if i not in calculated]
            old_ids_idx = [ids.index(i) for i in calculated if i in ids]
            calculated += new_ids

            new_ids_idx = [ids.index(i) for i in new_ids]

            p.plot(idx1, energies[idx1],
                   'ro', label='calculated')

            for new_id in new_ids:
                j = ids.index(new_id)
                old_id = self.DB.ase_db.get(id=new_id).initial_id
                if len(ids_old) > 0:
                    index_old = ids_old.index(old_id)
                    p.plot([j], [energies_old[index_old]], 'rx')
                    p.plot([j, j], [energies_old[index_old],
                                    energies[j]], 'r-')

            p.text(400, -2.4, '{} calculations'.format(len(self.train_ids)),
                   fontsize=14)
            p.xlabel('Structure id', fontsize=18)
            p.title('Predictions for batch {}'.format(prediction['batch_no']),
                    fontsize=20)

            p.ylabel('Energy(eV)', fontsize=18)
            p.ylim(-5, 5)
            p.legend()

            p.savefig('prediction_{}_batch_{}.pdf'.format(
                method, prediction['batch_no']))

            p.title('')

            good_energies = {'Anatase': -2.420,
                             'Rutile': -2.412,
                             'Rutile1': -2.411,
                             'Rutile2': -2.410,                                                                 'Rutile3': -2.388,
                             'Rutile4': -2.375,
                             'porous tetragon1': -2.3218,
                             'porous tetragon2': -2.3218,
                             'porous tetragon3': -2.31605,
                             'layered1': -2.281,
                             'layered1': -2.280,

                             }

            for key, value in good_energies.items():
                found_energies = np.isclose(value, energies[idx1], atol=0.001)
                if np.any(found_energies):
                    idx2 = np.where(found_energies)[0][0]
                    print(idx2)
                    p.text(idx1[idx2]+0.1, energies[idx1][idx2]-0.01, key)

            p.savefig('zoom_prediction_{}_batch_{}.pdf'.
                      format(method,
                             prediction['batch_no']))
            unc_old = unc.copy()
            energies_old = energies.copy()
            ids_old = ids.copy()

        p.show()

    def test_model(self, method='UCB', kappa=None, plot=True):

        lowest_energy = list(self.DB.ase_db.select(relaxed=1, sort='Ef'))[0].Ef
        calculated = []
        i = 0

        from matplotlib import cm
        colors = cm.get_cmap('viridis')(np.linspace(0, 1, 20))
        for prediction in self.test_run(acquire=method, kappa=kappa):
            i += 1
            idx = np.argsort(prediction['energies'])
            energies = prediction['energies'][idx]
            unc = prediction['vars'][idx]
            ids = prediction['ids'][idx]

            idx1 = np.where(unc == 0)[0]
            print(ids[idx1])
            n_calc = len(idx1)

            if plot:
                p.figure()
                p.plot(range(len(idx)), energies, '-', color=colors[i],
                       label='batch no {}, {} calculations'.format(str(i), n_calc))
                p.plot(idx1, energies[idx1], 's', color=colors[i],
                       fillstyle='none')

            if plot:
                p.xlabel('Structure id')
                p.ylabel('Energy(eV)')
                titlestr = '{} aquisition'.format(method)
                if kappa:
                    titlestr += ' kappa={}'.format(kappa)
                p.title(titlestr)
                p.legend()
            if i > 8:
                break
        p.show()

        return i

    def test_prototype_change(self):
        count_all = 0
        count_changed = 0
        for row in self.DB.ase_db.select(relaxed=1):
            count_all += 1
            p_name = row.p_name
            row_i = self.DB.ase_db.get(id=row.initial_id)
            p_name_i = row_i.p_name
            if not p_name == p_name_i:
                count_changed += 1
                print(row_i.id, '-->', row.id,
                      '{} --> {}'.format(p_name, p_name_i))
        print('{} out of {} structures changed symmetry during optimization'
              .format(count_changed, count_all))

    def plot_performance(self, verbose=False):
        train_ids = self.DB.get_completed_structure_ids()
        p_names = []
        energies = []
        for i in train_ids:
            entry = self.DB.ase_db.get(i)
            p_names += [entry.p_name]
            energies += [entry.energy]

        p_names, index = np.unique(p_names, return_index=True)
        train_ids = np.array(train_ids)[index]
        energies = np.array(energies)[index]

        remove_idx = []
        for i, e in enumerate(energies):
            if np.any(np.isclose(e, energies[i+1:], atol=0.01)):
                remove_idx += [i]

        train_ids = np.delete(train_ids, remove_idx)

        errors = []
        se = []
        errors_nonrelaxed = []
        se_nr = []
        energies = []
        mine = 100
        maxe = -100

        p.figure(figsize=(8, 8))
        print('Using {} training points'.format(len(train_ids) - 1))
        print('-------------------------------')
        for j in range(len(train_ids)):
            self.test_ids = [train_ids[j]]
            self.train_ids = np.delete(train_ids, j)
            initial_id = self.DB.ase_db.get(
                id=int(self.test_ids[0])).initial_id
            self.test_ids += [initial_id]
            self.test_ids = sorted(self.test_ids)
            print(self.test_ids)
            row = self.DB.ase_db.get(id=int(self.test_ids[-1]))
            row_initial = self.DB.ase_db.get(id=initial_id)
            energy = row.Ef

            self.train_ml()

            errors += [abs(self.energies[1] - energy)]
            se += [(self.energies[1] - energy)**2]
            errors_nonrelaxed += [abs(self.energies[0] - energy)]
            se_nr += [(self.energies[0] - energy)**2]

            if np.any(self.energies < mine):
                mine = np.min(self.energies)
            if np.any(self.energies > maxe):
                maxe = np.max(self.energies)

            p.plot([energy, energy], self.energies, linestyle='--',
                   color='gray', zorder=0)
            p.plot([energy], [self.energies[0]], 'ro')
            p.plot([energy], [self.energies[-1]], 'bo')

            if verbose:
                print('--------------------------')
                print('ids: ', self.test_ids)
                print('Ef: ', energy)
                print('Prototype: ', row.p_name)
                print('fmax: ', row.fmax)
                print('error relaxed: ', self.energies[1] - energy)
                print('error nonrelaxed: ', self.energies[0] - energy)

        print('-------------------------------')
        print('MAE relaxed Voronoi: ', np.mean(errors))
        print('MAE non-relaxed Voronoi: ', np.mean(errors_nonrelaxed))
        print('MSE relaxed Voronoi: ', np.mean(se))
        print('MSE non-relaxed Voronoi: ', np.mean(se_nr))

        p.plot([100], [100], 'ro',
               label='pre-relaxation voronoi. MAE: {0:.3f}'.\
               format(np.mean(errors_nonrelaxed)))
        p.plot([100], [100], 'bo',
               label='post-relaxed voronoi. MAE: {0:.3f}'.\
               format(np.mean(errors)))

        p.plot([mine-10, maxe+10], [mine-10, maxe+10], 'k-')
        p.xlim(mine-0.1, maxe+0.1)
        p.ylim(mine-0.1, maxe+0.1)

        p.title('ML performance for n_train={}'.format(len(train_ids) - 1),
                fontsize=24)
        p.xlabel('DFT Energy/atom (eV)', fontsize=18)
        p.ylabel('ML Energy/atom (eV)', fontsize=18)

        p.legend()

        p.savefig('performance.pdf')
        p.show()


if __name__ == "__main__":
    MA = MetaAnalysis(chemical_formulas=['Cu2O'], max_atoms=10, batch_size=10)
    MA.plot_performance()
    MA.plot_predictions()
    MA.plot_test_model()
    MA.plot_acquisition()
    MA.test_prototype_change()



from protosearch.build_bulk.build_bulk import BuildBulk
from ase.visualize import view

BB = BuildBulk(47, ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'], ['Cu', 'Pd', 'Pt','Au', 'Cu', 'Pd', 'Pt','Au'])

atoms = BB.get_atoms(proximity=1)

view(atoms)



import numpy as np
import json
import string
from ase import Atoms
from ase.db import connect
from ase.formula import Formula
from ase.symbols import string2symbols
import pandas as pd

from protosearch.utils.data import metal_numbers
from protosearch.workflow.prototype_db import PrototypeSQL
from protosearch import build_bulk
from .fitness_function import get_connections
from .spglib_interface import SpglibInterface
from .cell_parameters import CellParameters


path = build_bulk.__path__[0]


class OqmdInterface:
    """Interace to OQMD entries to create structurally unique atoms objects"""

    def __init__(self, source='icsd'):
        if source == 'icsd':
            self.dbfile = path + '/oqmd_icsd.db'
        elif source == 'oqmd':
            self.dbfile = path + '/oqmd_ver3.db'

    def create_proto_data_set(self,
                              chemical_formula,
                              max_atoms=None):
        """Create a dataset of unique prototype structures.

        Creates a unique set of structures with the same stoicheometry and
        composition by substituting the desired elemetns into the dataset of
        unique OQMD structure prototypes.

        Parameters
        ----------
        chemical_formula: str
            desired chemical formula with elements inserted (ex. 'Al2O3')

        max_atoms: int
          maximum number of atoms in the primitive unit cell
        """

        distinct_protonames = \
            self.get_distinct_prototypes(
                chemical_formula=chemical_formula,
                max_atoms=max_atoms)

        data_list = []

        for proto_name in distinct_protonames:
            data_list += self.get_atoms_for_prototype(chemical_formula,
                                                      proto_name)

        df = pd.DataFrame(data_list)

        return df

    def store_enumeration(self, filename, chemical_formula, max_atoms=None):
        """ Saves the enumerated prototypes and atomic species in a database"""

        distinct_protonames = \
            self.get_distinct_prototypes(
                chemical_formula=chemical_formula,
                max_atoms=max_atoms)

        for count, proto_name in enumerate(distinct_protonames):
            data_list = self.get_atoms_for_prototype(chemical_formula,
                                                     proto_name)

            DB = PrototypeSQL(filename=filename)

            for d in data_list:
                atoms = d.pop('atoms')

                formula = d.pop('chemical_formula')
                original_formula = d.pop('original_formula')

                # Same format as Enumerator output
                entry = d.copy()
                entry['spaceGroupNumber'] = entry.pop('spacegroup')
                entry['name'] = entry.pop('p_name')
                entry['parameters'] = {}
                entry['source'] = 'icsd'

                structure_name = d['structure_name']

                # Save prototype
                DB.write_prototype(entry=entry)

                if DB.ase_db.count(structure_name=structure_name) > 0:
                    continue

                key_value_pairs = \
                    {'p_name': d['p_name'],
                     'spacegroup': d['spacegroup'],
                     'permutations': json.dumps(d['specie_permutations']),
                     'wyckoffs': json.dumps(d['wyckoffs']),
                     'species': json.dumps(d['species']),
                     'structure_name': structure_name,
                     'relaxed': 0,
                     'completed': 0,
                     'submitted': 0}

                # Save structure
                DB.ase_db.write(atoms, key_value_pairs)

            if (count + 1) % 10 == 0:
                print('  {}/{} completed'.format(count+1,
                                                 len(distinct_protonames)))

    def get_atoms_for_prototype(self,
                                chemical_formula,
                                proto_name,
                                fix_metal_ratio=False,
                                must_contain_nonmetal=False,
                                max_candidates=1):

        oqmd_db = connect(self.dbfile)

        atoms0 = Atoms(chemical_formula)
        numbers0 = atoms0.numbers
        symbols0 = atoms0.get_chemical_symbols()

        metal_count = len([n for n in numbers0 if n in metal_numbers])
        metal_ratio_0 = metal_count / len(numbers0)
        nonmetals_idx = [i for i, n in enumerate(numbers0)
                         if not n in metal_numbers]
        nonmetals_symbols = ''.join(np.array(symbols0)[nonmetals_idx])

        if must_contain_nonmetal:
            structures = list(oqmd_db.select(nonmetals_symbols,
                                             proto_name=proto_name))
        else:
            structures = list(oqmd_db.select(proto_name=proto_name))

        if len(structures) == 0:
            return []

        formulas = [Formula(s.formula).reduce()[0].format('metal')
                    for s in structures]
        chemical_formula = Formula(chemical_formula).format('metal')
        if chemical_formula in formulas:
            idx = formulas.index(chemical_formula)
            structures = structures[idx:] + structures[:idx]

        atoms_data = []

        for s in structures:
            atoms = s.toatoms()
            old_species = np.array(s.data['species'], dtype='U2')
            old_symbols = atoms.get_chemical_symbols()
            orig_formula = Formula(atoms.get_chemical_formula()).reduce()[
                0].format('metal')
            metal_count = len([n for n in atoms.numbers if n in metal_numbers])
            metal_ratio = metal_count / len(atoms)
            if fix_metal_ratio and not metal_ratio == metal_ratio_0:
                continue

            atoms_sub_list = self.substitute_atoms(atoms, symbols0)
            graphs = []
            for atoms in atoms_sub_list:
                new_symbols = atoms.get_chemical_symbols()
                new_species = old_species.copy()
                for i in range(len(new_symbols)):
                    os = old_symbols[i]
                    idx = [i for i, o in enumerate(old_species) if o == os]
                    if idx:
                        new_species[idx] = [new_symbols[i]] * len(idx)

                graphs += [get_connections(atoms, decimals=1)]
                # Get spacegroup and conventional atoms
                SPG = SpglibInterface(atoms)
                atoms = SPG.get_conventional_atoms()
                spacegroup = SPG.get_spacegroup()

                structure_name = str(spacegroup)
                for spec, wy_spec in zip(new_species, s.data['wyckoffs']):
                    structure_name += '_{}_{}'.format(spec, wy_spec)

                # Set new lattice constants
                CP = CellParameters(spacegroup=spacegroup)
                atoms = CP.optimize_lattice_constants(atoms)

                # Get primitive atoms
                atoms = SPG.get_primitive_atoms(atoms)

                atoms_data += [{'spacegroup': spacegroup,
                                'wyckoffs': s.data['wyckoffs'],
                                'species': list(new_species),
                                'structure_name': structure_name,
                                'natom': len(atoms),
                                'specie_permutations': s.data['permutations'],
                                'p_name': proto_name,
                                'chemical_formula': chemical_formula,
                                'original_formula': orig_formula,
                                'atoms': atoms.copy()}]

            if len(atoms_data) >= max_candidates:
                break


        indices = [i for i in range(len(atoms_data))
                   if not np.any(graphs[i] in graphs[:i])]

        graphs = [graphs[i] for i in indices]
        atoms_data = [atoms_data[i] for i in indices]

        return atoms_data

    def substitute_atoms(self, atoms, new_symbols):
        """ Substitute new elements into atoms object"""

        formula = atoms.get_chemical_formula()
        rep = Formula(formula).reduce()[1]

        chemical_symbols = np.array(atoms.get_chemical_symbols(), dtype='U2')

        unique_old, counts_old = np.unique(
            chemical_symbols, return_counts=True)
        counts_old = counts_old / rep

        idx = np.argsort(counts_old)
        counts_old = counts_old[idx]
        unique_old = unique_old[idx]
        unique_new, counts_new = np.unique(new_symbols, return_counts=True)

        idx = np.argsort(counts_new)
        counts_new = counts_new[idx]
        unique_new = unique_new[idx]

        new_perm_temp = [[]]

        for i, c in enumerate(counts_new):
            perm_temp = []
            same_count = np.where(counts_new == c)[0]

            new_perm = []
            for temp in new_perm_temp:
                for i2 in same_count:
                    sym = unique_new[i2]
                    if not sym in temp:
                        new_perm += [temp + [sym]]

            new_perm_temp = new_perm.copy()

        old_symbols = chemical_symbols.copy()

        atoms_list = []
        for unique_new in new_perm:
            atoms_temp = atoms.copy()

            for i, old_s in enumerate(unique_old):
                loc_symbols = [i for i, s in enumerate(
                    old_symbols) if s == old_s]
                chemical_symbols[loc_symbols] = np.repeat(
                    [unique_new[i]], len(loc_symbols))

            atoms_temp.set_chemical_symbols(chemical_symbols)
            atoms_list += [atoms_temp.copy()]

        return atoms_list

    def ase_db(self):
        self.db = connect(self.dbfile)

    def get_distinct_prototypes(self,
                                chemical_formula=None,
                                max_atoms=None):
        """ Get list of distinct prototype strings given certain filters.

        Parameters
        ----------
        formula: str
          stiochiometry of the compound, f.ex. 'AB2' or AB2C3
        repetition: int
          repetition of the stiochiometry
        """
        db = connect(self.dbfile)
        con = db.connection or db._connect()
        cur = con.cursor()

        sql_comm = \
            "select distinct value from text_key_values where key='proto_name'"

        if chemical_formula is None:
            cur.execute(sql_comm)
            prototypes_rep = cur.fetchall()
            prototypes += [p[0] for p in prototypes_rep]

            return prototypes

        # Transform formula to general 'AB2C3' format
        formula, elements = get_stoich_formula(chemical_formula)

        repetitions = [0]
        if max_atoms is not None:
            natoms_formula = len(Atoms(chemical_formula))
            r = 1
            while (repetitions[-1] + 1) * natoms_formula <= max_atoms:
                repetitions += [r]
                r += 1
            del repetitions[0]

        prototypes = []
        for r in repetitions:
            if r == 0:  # all repetitions
                match_formula = formula
            else:
                match_formula = formula + '\_{}'.format(r)
            full_sql_comm = sql_comm + \
                " and value like '{}\_%' ESCAPE '\\'".format(match_formula)

            cur.execute(full_sql_comm)

            prototypes_rep = cur.fetchall()
            prototypes += [p[0] for p in prototypes_rep]

        return prototypes


def get_stoich_formula(formula, return_elements=True):
    """Get element independent stoichiometry formula
    i.e. TiO2 -> AB2"""
    alphabet = list(string.ascii_uppercase)
    elem_list = string2symbols(formula)

    unique_symbols, counts = np.unique(elem_list, return_counts=True)

    sorted_idx = np.argsort(counts)

    formula = ''
    for i, j in enumerate(sorted_idx):
        if counts[j] > 1:
            formula += alphabet[i] + str(counts[j])
        else:
            formula += alphabet[i]
    if return_elements:
        return formula, unique_symbols
    else:
        return formula



import sys
import copy
import numpy as np
import pandas as pd

from catlearn.fingerprint.voro import VoronoiFingerprintGenerator
from catlearn.preprocess.clean_data import (
    clean_infinite,
    clean_variance,
    clean_skewness,
)

from catlearn.preprocess.scaling import standardize

# Featurizing methods implemented, must have corresponding class
feature_methods_dict = {
    "voronoi": "VoronoiFingerprint",
}


class FingerPrint:
    """General fingerprinting/featurizing class.

    Development Notes:
        * If the fingerprinting needs to know what the entire set of atoms
        that must be taken into account
        * Incorporate feature engineering
        * TEMP
    """

    def __init__(self,
                 feature_methods=None,
                 input_data=None,
                 input_index=None):
        """
        Parameters
        ----------
        feature_methods: list
            List of featurizing methods to apply to inputs.
        input_data: list or pandas_dataframe
            Pandas dataframe containing a column corresponding to the input
            objects to be featurized.
        input_index: str
            must be defined to index the correct column in inpud_data
        clean: Bool
            whether to clean features or not
        """
        self.feature_methods = feature_methods
        self.input_index = input_index
        self.input_data = input_data

        self.__check_class_inputs__()

        self.__instantiate_feature_classes__()

    def __check_class_inputs__(self):
        """
        checking_class_inputs
        """

        # feature_methods
        err_mess = "feature_methods must be of type <list>"
        assert isinstance(self.feature_methods, list), err_mess

        for method in self.feature_methods:
            err_mess_i = "feature methods must be in the following: \n"
            err_mess_i += str(feature_methods_dict.keys())
            assert method in feature_methods_dict.keys(), err_mess_i

        # input_data
        is_pd_df = isinstance(self.input_data, pd.DataFrame)
        if not is_pd_df:
            msg = "Please give input_data as a pandas dataframe \n"
            msg += "At the least a dataframe with 1 column"
            raise TypeError("Please give input_data as a pandas dataframe")

        # input_index
        # Is this necessary?
        if isinstance(self.input_index, list):
            pass
            # self.input_index = tuple(input_index)
        elif isinstance(self.input_index, set):
            pass

        else:
            raise TypeError("input_index must be given as a <list> or <set>")

    def __instantiate_feature_classes__(self):
        out_dict = {}
        for method in self.feature_methods:
            feature_class_name = feature_methods_dict[method]
            class_i = getattr(sys.modules[__name__], feature_class_name)

            # COMBAK There were some issues caused here by the .loc method
            # returning either a pandas.DataFrame or a pandas.series
            # input_array = input_data.loc[:, input_index].tolist()
            input_array = \
                self.input_data.loc[:, self.input_index].iloc[:, 0].tolist()
            instance = class_i(input_array)
            out_dict[method] = instance

        self.feature_instances = out_dict

    def generate_fingerprints(self):
        # generate_fingerprints
        feature_instances = self.feature_instances
        input_data = self.input_data

        # Collecting fingerprint dataframes from fingerprint instances
        fingerprints = {}
        for name_i, feature_instance_i in feature_instances.items():
            feature_instance_i.generate_fingerprints()

            features_i = feature_instance_i.features

            # Checking type of fingerprints (must be pandas dataframe)
            # Fingerprints must be given as a pandas dataframe
            is_pd_df = isinstance(features_i, pd.DataFrame)
            err_mess_i = "Fingerprint class must return a pandas dataframe"
            assert is_pd_df, err_mess_i

            features_i = features_i.set_index(
                input_data.index,
                # np.array(rand_ids),
                drop=False, append=False,
                inplace=False, verify_integrity=True)

            fingerprints[name_i] = features_i

        fingerprints_out = pd.concat(fingerprints.values(),
                                     axis=1,
                                     keys=fingerprints.keys())\

        self.fingerprints = fingerprints_out
        return self.fingerprints

    def join_input_to_fingerprints(self):
        """Concancotate
        """
        # join_input_to_fingerprints
        input_data = self.input_data
        fingerprints = self.fingerprints

        df_out = pd.merge(input_data, fingerprints,
                          left_index=True,
                          right_index=True,
                          indicator=True,  # This was breaking the method for some reason
                          )

        # TODO
        # Check that operation was succesful
        # Merge command shouldn't be dropping any rows

        # print(len(input_data))
        # print(len(fingerprints))
        # print(len(df_out))

        if len(input_data) != len(fingerprints):
            print("MISTAKE iasdjfisj")
        if len(input_data) != len(df_out):
            print("MISTAKE iasdjfisj2")
        if len(fingerprints) != len(df_out):
            print("MISTAKE iasdjfisj3")

        self.fingerprints = df_out

        # return(input_data, fingerprints)
        # return(df_out)


class VoronoiFingerprint:
    """
    Uses the CatLearn interface to MagPie Voronoi Fingerprint

    """
    from catlearn.fingerprint.voro import VoronoiFingerprintGenerator

    def __init__(self,
                 atoms_list):
        """Voronoi fingerprinting setup.

        Parameters
        ----------
        atoms_list : list
            A list of ase atoms objects to be featurized.


        Development Notes:
        ------------------
        * This featurizing method is specific to atoms objects, for the sake
        of generality lets just assume that inputs can be anything, therefore
        check that inputs are atoms objects

        """
        self.check_inputs(atoms_list)
        self.Voro_inst = VoronoiFingerprintGenerator(
            atoms_list,
            delete_temp=False)

    def check_inputs(self, atoms_list):
        # check atoms list
        from ase import Atoms

        type_check_list = [isinstance(atoms_i, Atoms)
                           for atoms_i in atoms_list]
        err_mess = "Inputs to Voronoi must be atom objects"
        assert all(type_check_list), err_mess

    def generate_fingerprints(self):
        self.features = self.Voro_inst.generate()

    def get_fingerprints(self):
        self.generate_features()
        return self.features


def clean_features(features, scale=False):
    remove_indices = {'train': np.array([], dtype=int),
                      'test': np.array([], dtype=int)}
    for key, feature_set in features.items():
        if feature_set is None or len(feature_set) == 0:
            continue
        bad_structure_indices = \
            np.where(np.isfinite(feature_set).all(axis=1) == False)
        for b in bad_structure_indices:
            if len(np.where(np.isfinite(feature_set[b]) == False)) > 1:
                remove_indices[key] = np.append(remove_indices[key], b)

        features[key] = np.delete(feature_set, remove_indices[key], axis=0)

    if not 'test' in features:
        features['test'] = None

    # Finite features
    features = clean_infinite(features['train'],
                              features['test'])

    # Clean variance
    features = clean_variance(features['train'],
                              features['test'])

    # Clean skewness & standardize
    if features['test'] is None or len(features['test']) == 0:
        features = clean_skewness(features['train'],
                                  skewness=3)
        if scale:
            features = standardize(features['train'])

    else:
        features = clean_skewness(features['train'],
                                  features['test'],
                                  skewness=3)
        if scale:
            features = standardize(features['train'],
                                   features['test'])

    return features, remove_indices



import os

from .cell_parameters import CellParameters


class BuildBulk(CellParameters):
    """
    Set up bulk structures for a given prototype specification.

    Prototypes can be enumeated with the Bulk prototype enumerator,
    developed by A. Jain described in:
    A. Jain and T. Bligaard, Phys. Rev. B 98, 214112 (2018)

    Parameters:

    spacegroup: int
        int between 1 and 230
    wyckoffs: list
        wyckoff positions, for example ['a', 'a', 'b', 'c']
    species: list
        atomic species, for example ['Fe', 'O', 'O', 'O']
    """

    def __init__(self,
                 spacegroup,
                 wyckoffs,
                 species
                 ):

        super().__init__(spacegroup=spacegroup,
                         wyckoffs=wyckoffs,
                         species=species)

        assert (0 < spacegroup < 231 and isinstance(spacegroup, int)), \
            'Spacegroup must be an integer between 1 and 230'

        self.spacegroup = spacegroup
        self.wyckoffs = wyckoffs
        self.species = species

    def get_atoms(self, proximity=1, cell_parameters=None, primitive_cell=True):
        """Get one atoms object pr prototype

        Parameters:

        proximity: float
            number close to 1, specifying the proximity of atoms in the 
            hard sphere model. r_spheres = proximity * r_covalent

        cell_parameters: dict
            Optional specification of cell parameters, 
             such as {'a': 3.7, 'alpha': 75}.
             Otherwise a guees for parameters will be provided by the
             CellParameters module."""

        master_parameters = cell_parameters or {}
        cell_parameters = self.get_parameter_estimate(master_parameters=master_parameters,
                                                      proximity=proximity,
                                                      max_candidates=1)

        return self.construct_atoms(cell_parameters[0],
                                    primitive_cell=primitive_cell)

    def get_wyckoff_candidate_atoms(self,
                                    proximity=1,
                                    cell_parameters=None,
                                    primitive_cell=True,
                                    return_parameters=False,
                                    max_candidates=None):
        """Returns a list of atomic structures with different wyckoff settings"""
        master_parameters = cell_parameters or {}

        cell_parameters = \
            self.get_parameter_estimate(master_parameters=master_parameters,
                                        proximity=proximity,
                                        max_candidates=max_candidates)

        atoms_list = []
        for c_p in cell_parameters:
            atoms_list += [self.construct_atoms(c_p,
                                                primitive_cell=primitive_cell)]

        if return_parameters:
            return atoms_list, cell_parameters

        return atoms_list



__maintainer__ = "Meng Zhao"

import pprint
import csv
import json
import os,sys
import ase
from ase.db import connect
from ase.visualize import view

from protosearch.build_bulk.build_bulk import BuildBulk


database = connect('structures.db')
file_path = 'AB2_unique.csv'

csv_file = open(file_path)
data_complete = csv.reader(csv_file, delimiter=',')

# generate POSCAR scripts for bulk structures 
for i, row in enumerate(data_complete):
    if i < 1:
        continue

    print('---------- Structure {} ---------------'.format(i))

    cell_parameters = json.loads(row[2].replace("'", '"'))

    # Allow lattice constants to be optimized
    for del_param in ['a', 'b', 'c']:
        if del_param in cell_parameters:
            del cell_parameters[del_param]

    
    spacegroup = int(row[4])
    wyckoffs = eval(row[6])
    species = eval(row[5])
    atomA = species[0]

    # Change to TiO2 structure
    for i in range(len(species)):
        if species[i] == atomA:
            species[i] = 'Ti'
        else:
            species[i] = 'O'
    try: 
        bulk_generator = BuildBulk(spacegroup = int(row[4]),
                                   wyckoffs = eval(row[6]),
                                   species = species,
                                   cell_parameters=cell_parameters)
    
        atoms = bulk_generator.atoms
        
        if atoms:
            view(atoms)
            database.write(atoms)
    except BaseException:        
        print('Error: something is wrong with this structure')
        continue
